import collections
from typing import Optional
import typing

import mincepy
from . import depositors
from . import refs
from . import types

__all__ = 'BaseSavableObject', 'ConvenienceMixin', 'SimpleSavable', 'AsRef'

AttrSpec = collections.namedtuple('AttrSpec', 'name as_ref')


def AsRef(name: str) -> AttrSpec:  # pylint: disable=invalid-name
    """Create an attribute specification for an attribute that should be stored by reference"""
    return AttrSpec(name, True)


class BaseSavableObject(types.SavableObject):
    """A helper class that makes a class compatible with the historian by flagging certain
    attributes which will be saved/loaded/hashed and compared in __eq__.  This should be an
    exhaustive list of all the attributes that define this class.  If more complex functionality
    is needed then the standard SavableComparable interface methods should be overwritten."""
    ATTRS = tuple()
    IGNORE_MISSING = True  # When loading ignore attributes that are missing in the record

    def __new__(cls, *_args, **_kwargs):
        new_instance = super(BaseSavableObject, cls).__new__(cls)
        attrs = {}
        for entry in cls.mro():
            try:
                class_attrs = getattr(entry, 'ATTRS')
            except AttributeError:
                pass
            else:
                for attr_spec in class_attrs:
                    if isinstance(attr_spec, str):
                        # If it's just a string then default to store by value
                        attr_spec = AttrSpec(attr_spec, False)

                    # Check that it's not already there so higher up in the MRO is always kept
                    if attr_spec.name not in attrs:
                        attrs[attr_spec.name] = attr_spec

        setattr(new_instance, '__attrs', attrs.values())
        return new_instance

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False

        return all(
            getattr(self, attr.name) == getattr(other, attr.name) for attr in self.__get_attrs())

    def yield_hashables(self, hasher):
        yield from hasher.yield_hashables([getattr(self, attr.name) for attr in self.__get_attrs()])

    def save_instance_state(self, _saver) -> dict:
        saved_state = {}
        for attr in self.__get_attrs():
            item = getattr(self, attr.name)
            if attr.as_ref:
                item = refs.ObjRef(item)
            saved_state[attr.name] = item

        super().save_instance_state(_saver)
        return saved_state

    def load_instance_state(self, saved_state, _loader):
        super(BaseSavableObject, self).load_instance_state(saved_state, _loader)
        for attr in self.__get_attrs():
            try:
                obj = saved_state[attr.name]
            except KeyError:
                if self.IGNORE_MISSING:
                    # Set any missing attributes to None
                    setattr(self, attr.name, None)
                else:
                    raise
            else:
                if attr.as_ref:
                    if obj:
                        obj = obj()
                    else:
                        obj = None
                setattr(self, attr.name, obj)

    def __get_attrs(self) -> typing.Sequence[AttrSpec]:
        return getattr(self, '__attrs')


class ConvenienceMixin:
    """A mixin that adds convenience methods to your savable object"""

    @property
    def obj_id(self):
        if self._historian is None:
            return None
        return self._historian.get_obj_id(self)

    def get_meta(self) -> Optional[dict]:
        """Get the metadata dictionary for this object"""
        if self._historian is None:
            return None
        return self._historian.meta.get(self)

    def set_meta(self, meta: Optional[dict]):
        """Set the metadata dictionary for this object"""
        if self._historian is None:
            raise RuntimeError("Object must be saved before the metadata can be set")

        self._historian.meta.set(self, meta)

    def update_meta(self, meta: dict):
        """Update the metadata dictionary for this object"""
        if self._historian is None:
            raise RuntimeError("Object must be saved before the metadata can be updated")

        self._historian.meta.update(self, meta)

    def is_saved(self) -> bool:
        """Returns True if this object is saved with a historian"""
        if self._historian is not None:
            return self._historian.is_saved(self)

        return False

    def save(self, meta: dict = None):
        """Save the object"""
        historian = self._historian or mincepy.get_historian()
        return historian.save_one(self, meta=meta)

    def sync(self):
        """Update the state of this object by loading the latest version from the historian"""
        if self._historian is not None:
            self._historian.sync(self)

    def save_instance_state(self, saver: depositors.Saver):
        self._on_save(saver)
        return super().save_instance_state(saver)

    def load_instance_state(self, saved_state, loader: 'mincepy.Loader'):
        """Take the given object and load the instance state into it"""
        super().load_instance_state(saved_state, loader)
        self._on_load(loader)

    def _on_save(self, saver: depositors.Saver):
        # Check if we've been assigned an object id, otherwise we're just being saved by value
        if saver.get_historian().get_obj_id(self) is not None:
            self._historian = saver.get_historian()

    def _on_load(self, loader: depositors.Loader):
        if loader.get_historian().get_obj_id(self) is not None:
            self._historian = loader.get_historian()


class SimpleSavable(ConvenienceMixin, BaseSavableObject):
    """A BaseSavableObject with convenience methods mixed in"""



from typing import Any, Optional

import pytray.pretty

import mincepy  # pylint: disable=unused-import

__all__ = 'ObjectMigration', 'ObjectMigrationMeta'


class ObjectMigrationMeta(type):
    PARENT = None

    def __init__(cls, name, bases, dct):
        super().__init__(name, bases, dct)

        if cls.PARENT is None or cls == cls.PARENT:
            # This one is the parent class for others and doesn't need to pass the checks
            return

        if cls.VERSION is None:
            raise RuntimeError("Migration version not set")

        if cls.PREVIOUS is not None and cls.VERSION <= cls.PREVIOUS.VERSION:
            raise RuntimeError(
                "A migration must have a version number higher than the previous migration.  {}"
                ".VERSION is {} while {}.VERSION is {}".format(
                    pytray.pretty.type_string(cls.PREVIOUS), cls.PREVIOUS.VERSION,
                    pytray.pretty.type_string(cls), cls.VERSION))

        if cls.NAME is None:
            cls.NAME = cls.__class__.__name__


class ObjectMigration(metaclass=ObjectMigrationMeta):
    NAME = None  # type: Optional[str]
    VERSION = None  # type: Optional[int]
    PREVIOUS = None  # type: Optional[ObjectMigration]

    @classmethod
    def upgrade(cls, saved_state, loader: 'mincepy.Loader') -> Any:
        """
        This method should take the saved state, which will have been created with the previous
        version, and return a new saved state that is compatible with this version.

        :raises mincepy.MigrationError: if a problem is encountered during migration
        """
        raise NotImplementedError


ObjectMigrationMeta.PARENT = ObjectMigration



import contextlib
import copy
from typing import MutableMapping, Any, List, Sequence, Optional, Dict, Set, Union, overload
import weakref

import deprecation

from . import archives
from . import exceptions
from . import operations
from . import records
from . import utils
from . import version as version_mod


class LiveObjects:
    """A container for storing live objects"""

    def __init__(self):
        # Live object -> data records
        self._records = \
            utils.WeakObjectIdDict()  # type: MutableMapping[object, archives.DataRecord]
        # Obj id -> (weak) object
        self._objects = weakref.WeakValueDictionary()  # type: MutableMapping[Any, object]

    def __str__(self):
        return "{} live".format(len(self._objects))

    def __contains__(self, item):
        """Determine if an object instance is in this live objects container"""
        return item in self._records

    def insert(self, obj, record: records.DataRecord):
        self._records[obj] = record
        self._objects[record.obj_id] = obj

    def update(self, live_objects: 'LiveObjects'):
        """Like a dictionary update, take the given live objects container and absorb it into
        ourselves overwriting any existing values and incorporating any new"""
        # pylint: disable=protected-access
        self._records.update(live_objects._records)
        self._objects.update(live_objects._objects)

    def remove(self, obj_id) -> Any:
        """Remove an object from the collection.  Returns the removed object.

        :raises: :class:`mincepy.NotFound` if the ID is not found
        """
        try:
            return self._records.pop(self._objects.pop(obj_id))
        except KeyError:
            raise exceptions.NotFound(obj_id)

    def get_record(self, obj) -> records.DataRecord:
        try:
            return self._records[obj]
        except KeyError:
            raise exceptions.NotFound("No live object found '{}'".format(obj))

    @overload
    def get_object(self, identifier: records.SnapshotId):  # pylint: disable=no-self-use
        ...

    @overload
    def get_object(self, identifier: Any):  # pylint: disable=no-self-use
        ...

    def get_object(self, identifier: Union[records.SnapshotId, Any]):
        """Get an object from the collection either by snapshot id or object id

        :raises: :class:`mincepy.NotFound` if the ID is not found
        """
        if isinstance(identifier, records.SnapshotId):
            for obj, record in self._records.items():
                if record.snapshot_id == identifier:
                    return obj
            raise exceptions.NotFound(identifier)

        # Must be an object id
        try:
            return self._objects[identifier]
        except KeyError:
            raise exceptions.NotFound("No live object with id '{}'".format(identifier))

    def get_snapshot_id(self, obj) -> records.SnapshotId:
        """Given an object, get the snapshot id"""
        try:
            return self._records[obj].snapshot_id
        except KeyError:
            raise exceptions.NotFound(obj)


class RollbackTransaction(Exception):
    pass


class Transaction:
    """A transaction object for keeping track of actions performed in a transaction that will be
    committed at the end.

    A transaction has no interaction with the database and simply stores a transient state.  For
    this reason things like constrains are not enforced and database queries will no reflect
    mutations performed within the transaction (only upon commit).
    """

    # pylint: disable=too-many-public-methods

    @deprecation.deprecated(deprecated_in="0.14.4",
                            removed_in="0.16.0",
                            current_version=version_mod.__version__,
                            details="Use get_snapshot_id_for_live_object() instead")
    def get_reference_for_live_object(self, obj) -> records.SnapshotId:
        return self.get_snapshot_id_for_live_object(obj)

    @deprecation.deprecated(deprecated_in="0.14.4",
                            removed_in="0.16.0",
                            current_version=version_mod.__version__,
                            details="Use get_live_object_from_snapshot_id() instead")
    def get_live_object_from_reference(self, snapshot_id: records.SnapshotId):
        return self.get_live_object(snapshot_id)

    @deprecation.deprecated(deprecated_in="0.14.4",
                            removed_in="0.16.0",
                            current_version=version_mod.__version__,
                            details="Use get_live_object() instead")
    def get_live_object_from_snapshot_id(self, snapshot_id: records.SnapshotId):
        return self.get_live_object(snapshot_id)

    def __init__(self):
        # Records staged for saving to the archive
        self._staged = []  # type: List[operations.Operation]

        self._deleted = set()  # A set of deleted obj ids

        self._live_objects = LiveObjects()
        # Snapshot id -> obj for objects currently being saved
        self._in_progress_cache = {}  # type: Dict[records.SnapshotId, object]

        # Snapshots: snapshot id -> obj
        self._snapshots = {}  # type: Dict[records.SnapshotId, Any]
        # Maps from object id -> metadata dictionary
        self._metas = {}  # type: Dict[Any, dict]

    def __str__(self):
        return "{}, {} live ref(s), {} snapshots, {} staged".format(self._live_objects,
                                                                    len(self._in_progress_cache),
                                                                    len(self._snapshots),
                                                                    len(self._staged))

    @property
    def live_objects(self) -> LiveObjects:
        return self._live_objects

    @property
    def deleted(self) -> Set:
        return self._deleted

    @property
    def snapshots(self):
        return self._snapshots

    @property
    def metas(self) -> dict:
        return self._metas

    # region LiveObjects

    def insert_live_object(self, obj, record: records.DataRecord):
        """Insert a live object along with an up-to-date record into the transaction"""
        if self.is_deleted(record.obj_id):
            raise ValueError("Object with id '{}' has already been deleted!".format(record.obj_id))

        sid = record.snapshot_id
        if sid in self._in_progress_cache:
            assert self._in_progress_cache[sid] is obj

        self._live_objects.insert(obj, record)

    @contextlib.contextmanager
    def prepare_for_saving(self, snapshot_id: records.SnapshotId, obj):
        """Insert a snapshot reference for an object into the transaction"""
        self._in_progress_cache[snapshot_id] = obj
        try:
            yield
        except Exception:  # Need this for the 'else' pylint: disable=try-except-raise
            raise
        else:
            if self._live_objects.get_snapshot_id(obj) != snapshot_id:
                raise RuntimeError(
                    "Problem saving object with snapshot id '{}', "
                    "the snapshot id saved does not match that given".format(snapshot_id))
        finally:
            del self._in_progress_cache[snapshot_id]

    @overload
    def get_live_object(self, identifier: records.SnapshotId) -> object:  # pylint: disable=no-self-use
        ...

    @overload
    def get_live_object(self, identifier: Any) -> object:  # pylint: disable=no-self-use
        ...

    def get_live_object(self, identifier: Union[records.SnapshotId, Any]) -> object:
        if isinstance(identifier, records.SnapshotId):
            self._ensure_not_deleted(identifier.obj_id)
            try:
                return self._in_progress_cache[identifier]
            except KeyError:
                return self._live_objects.get_object(identifier)

        # Must be an object id
        self._ensure_not_deleted(identifier)
        return self._live_objects.get_object(identifier)

    def get_record_for_live_object(self, obj) -> records.DataRecord:
        return self._live_objects.get_record(obj)

    def get_snapshot_id_for_live_object(self, obj) -> records.SnapshotId:
        for ref, cached_obj in self._in_progress_cache.items():
            if obj is cached_obj:
                return ref

        return self._live_objects.get_snapshot_id(obj)

    def delete(self, obj_id):
        """Mark an object as deleted"""
        try:
            self._live_objects.remove(obj_id)
        except exceptions.NotFound:
            pass

        self.set_meta(obj_id, None)
        self._deleted.add(obj_id)

    def is_deleted(self, obj_id):
        return obj_id in self.deleted

    # endregion LiveObjects

    # region meta

    def set_meta(self, obj_id, meta: Optional[dict]):
        """Set an object's metadata.  Can pass None to unset."""
        self._ensure_not_deleted(obj_id)
        self._metas[obj_id] = copy.deepcopy(meta)

    def get_meta(self, obj_id) -> dict:
        """Get an object's metadata.
        If the returned metadata is None then it means the user has set this metadata to be removed.

        :raise exceptions.NotFound: if not metadata information for the object is contained in this
            transaction.
        """
        self._ensure_not_deleted(obj_id)
        try:
            return self._metas[obj_id]
        except KeyError:
            raise exceptions.NotFound

    # endregion

    def insert_snapshot(self, obj, snapshot_id):
        self._snapshots[snapshot_id] = obj

    def get_snapshot(self, snapshot_id):
        try:
            return self._snapshots[snapshot_id]
        except KeyError:
            raise exceptions.NotFound("No snapshot with id '{}' found".format(snapshot_id))

    def stage(self, op: operations.Operation):  # pylint: disable=invalid-name
        """Stage an operation to be carried out on completion of this transaction"""
        self._staged.append(op)

    @property
    def staged(self) -> Sequence[operations.Operation]:
        """The list of records that were staged during this transaction"""
        return self._staged

    @contextlib.contextmanager
    def nested(self):
        nested = NestedTransaction(self)
        try:
            yield nested
        except RollbackTransaction:
            pass
        else:
            # Update our transaction with the nested
            self._update(nested)

    def _update(self, transaction: 'Transaction'):
        """Absorb a nested transaction into this one, done at the end of a nested context"""
        # pylint: disable=protected-access
        self._live_objects.update(transaction.live_objects)
        self._snapshots.update(transaction.snapshots)
        self._in_progress_cache.update(transaction._in_progress_cache)
        self._staged.extend(transaction.staged)
        self._metas.update(transaction.metas)
        for deleted in transaction.deleted:
            self.delete(deleted)

    @staticmethod
    def rollback():
        raise RollbackTransaction

    def _ensure_not_deleted(self, obj_id):
        """Make sure that an object id has not been deleted in this transaction.  Raises an
        ObjectDeleted exception if so."""
        if self.is_deleted(obj_id):
            raise exceptions.ObjectDeleted(obj_id)


class NestedTransaction(Transaction):

    def __init__(self, parent: Transaction):
        super(NestedTransaction, self).__init__()
        self._parent = parent

    def __str__(self):
        return "{} (parent: {})".format(super().__str__(), self._parent)

    def get_live_object(self, identifier) -> object:
        try:
            return super().get_live_object(identifier)
        except exceptions.NotFound:
            return self._parent.get_live_object(identifier)

    def get_snapshot_id_for_live_object(self, obj):
        try:
            return super().get_snapshot_id_for_live_object(obj)
        except exceptions.NotFound:
            return self._parent.get_snapshot_id_for_live_object(obj)

    def get_record_for_live_object(self, obj):
        try:
            return super().get_record_for_live_object(obj)
        except exceptions.NotFound:
            return self._parent.get_record_for_live_object(obj)

    def get_snapshot(self, snapshot_id):
        try:
            return super(NestedTransaction, self).get_snapshot(snapshot_id)
        except exceptions.NotFound:
            return self._parent.get_snapshot(snapshot_id)

    def get_meta(self, obj_id) -> dict:
        try:
            return super(NestedTransaction, self).get_meta(obj_id)
        except exceptions.NotFound:
            return self._parent.get_meta(obj_id)

    def is_deleted(self, obj_id):
        if obj_id in self.deleted:
            return True

        return self._parent.is_deleted(obj_id)



import functools

from mincepy import q
from . import db


def pipeline_latest_version(data_collection: str) -> list:
    """Returns a pipeline that will take the incoming data record documents and for each one find
    the latest version."""
    oid_var = '${}'.format(db.OBJ_ID)
    ver_var = '${}'.format(db.VERSION)

    pipeline = []
    pipeline.extend([
        # Group by object id the maximum version
        {
            '$group': {
                '_id': oid_var,
                'max_ver': {
                    '$max': ver_var
                }
            }
        },
        # Then do a lookup against the same collection to get the records
        {
            '$lookup': {
                'from': data_collection,
                'let': {
                    'obj_id': '$_id',
                    'max_ver': '$max_ver'
                },
                'pipeline': [{
                    '$match': {
                        '$expr':
                            q.and_(
                                # Match object id and version
                                q.eq_(oid_var, '$$obj_id'),
                                q.eq_(ver_var, '$$max_ver')),
                    }
                }],
                'as': 'latest'
            }
        },
        # Now unwind and promote the 'latest' field
        {
            '$unwind': {
                'path': '$latest'
            }
        },
        {
            '$replaceRoot': {
                'newRoot': '$latest'
            }
        },
    ])

    return pipeline


def pipeline_match_metadata(meta: dict, meta_collection: str, local_field: str):
    pipeline = []

    pipeline.append({
        '$lookup': {
            'from': meta_collection,
            'localField': local_field,
            'foreignField': '_id',
            'as': '_meta'
        }
    })
    # _meta should only contain at most one entry per document i.e. the metadata for
    # that object.  So check that for the search criteria
    pipeline.append({'$match': {'_meta.0.{}'.format(key): value for key, value in meta.items()}})

    return pipeline


class QueryBuilder:
    """Simple MongoDB query builder.  Creates a compound query of one or more more terms"""

    def __init__(self, *terms):
        self._terms = []
        for term in terms:
            self.and_(term)

    def and_(self, *term: dict):
        for entry in term:
            if not isinstance(entry, dict):
                raise TypeError(entry)

        self._terms.extend(term)

    def build(self):
        if not self._terms:
            return {}
        if len(self._terms) == 1:
            return self._terms[0]

        return {'$and': self._terms.copy()}


def flatten_filter(entry_name: str, query) -> list:
    """Expand nested search criteria, e.g. state={'color': 'red'} -> {'state.colour': 'red'}"""
    if isinstance(query, dict):
        transformed = _transform_query_keys(query, entry_name)
        flattened = [{key: value} for key, value in transformed.items()]
    else:
        flattened = [{entry_name: query}]

    return flattened


@functools.singledispatch
def _transform_query_keys(entry, prefix: str = ''):  # pylint: disable=unused-argument
    """Transform a query entry into the correct syntax given a global prefix and the entry itself"""
    return entry


@_transform_query_keys.register(list)
def _(entry: list, prefix: str = ''):
    return [_transform_query_keys(value, prefix) for value in entry]


@_transform_query_keys.register(dict)
def _(entry: dict, prefix: str = ''):
    transformed = {}
    for key, value in entry.items():
        if key.startswith('$'):
            if key in ('$and', '$not', '$nor', '$or'):
                transformed[key] = _transform_query_keys(value, prefix)
            else:
                update = {prefix: {key: value}} if prefix else {key: value}
                transformed.update(update)
        else:
            to_join = [prefix, key] if prefix else [key]
            # Don't pass the prefix on, we've consumed it here
            transformed['.'.join(to_join)] = _transform_query_keys(value)

    return transformed



"""This module contains various strategies for loading, saving and migrating objects in the archive
"""

from abc import ABCMeta, abstractmethod
import logging
from typing import Optional, MutableMapping, Any, Iterable, Sequence

import deprecation
from pytray import tree

import mincepy  # pylint: disable=unused-import
from . import archives
from . import exceptions
from . import operations
from . import records
from . import staging
from . import transactions  # pylint: disable=unused-import
from . import version as version_mod

__all__ = 'Saver', 'Loader', 'SnapshotLoader', 'LiveDepositor', 'Migrator'

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Base(metaclass=ABCMeta):
    """Common base for loader and saver"""

    def __init__(self, historian):
        self._historian = historian  # type: mincepy.Historian

    def get_archive(self) -> archives.Archive:
        return self._historian.archive

    def get_historian(self) -> 'mincepy.Historian':
        """
        Get the owning historian.

        :return: the historian
        """
        return self._historian


class Saver(Base, metaclass=ABCMeta):
    """A depositor that knows how to save records to the archive"""

    @deprecation.deprecated(deprecated_in="0.14.2",
                            removed_in="0.16.0",
                            current_version=version_mod.__version__,
                            details="Use get_snapshot_id() instead")
    def ref(self, obj) -> records.SnapshotId:
        """Get a persistent reference for the given object"""
        return self.get_snapshot_id(obj)

    @abstractmethod
    def get_snapshot_id(self, obj) -> records.SnapshotId:
        """Get a persistent reference for the given object"""

    def encode(self, obj, schema=None, path=()):
        """Encode a type for archiving"""
        historian = self.get_historian()

        if historian.is_primitive(obj):
            # Deal with the special containers by encoding their values if need be
            return tree.transform(self.encode, obj, path, schema=schema)

        # Store by value
        helper = historian.get_helper(type(obj), auto_register=True)
        save_state = helper.save_instance_state(obj, self)
        if not historian.is_primitive(save_state):
            raise RuntimeError("Saved state must be one of the primitive types")

        schema_entry = [path, helper.TYPE_ID]
        version = helper.get_version()
        if version is not None:
            schema_entry.append(version)

        if schema is not None:
            schema.append(schema_entry)

        return self.encode(save_state, schema, path)

    def save_state(self, obj):
        """Save the state of an object and return the encoded state ready to be archived in a
        record"""
        schema = []
        saved_state = self.encode(obj, schema)
        return {records.STATE: saved_state, records.STATE_TYPES: schema}


class Loader(Base, metaclass=ABCMeta):
    """A loader that knows how to load objects from the archive"""

    @abstractmethod
    def load(self, snapshot_id: records.SnapshotId):
        """Load an object"""

    def decode(self,
               encoded,
               schema: records.StateSchema = None,
               path=(),
               created_callback=None,
               updates=None):
        """Given the encoded state and an optional schema that defines the type of the encoded
        objects this method will decode the saved state and load the object."""
        try:
            entry = schema[path]
        except KeyError:
            return self._recursive_unpack(encoded, schema, path, created_callback)
        else:
            saved_state = encoded
            helper = self.get_historian().get_helper(entry.type_id)
            if helper.IMMUTABLE:
                saved_state = self._recursive_unpack(encoded, schema, path, created_callback)

            new_obj = helper.new(saved_state)
            if new_obj is None:
                raise RuntimeError("Helper '{}' failed to create a class given state '{}'".format(
                    helper.__class__, saved_state))

            if created_callback is not None:
                created_callback(path, new_obj)

            if not helper.IMMUTABLE:
                saved_state = self._recursive_unpack(encoded, schema, path, created_callback,
                                                     updates)

            updated = helper.ensure_up_to_date(saved_state, entry.version, self)
            if updated is not None:
                # Use the current version of the record
                saved_state = updated
                if updates is not None:
                    updates[path] = updated

            helper.load_instance_state(new_obj, saved_state, self)
            return new_obj

    def _recursive_unpack(self,
                          encoded_saved_state,
                          schema: records.StateSchema = None,
                          path=(),
                          created_callback=None,
                          updates=None):
        """Unpack a saved state expanding any contained objects"""
        return tree.transform(self.decode,
                              encoded_saved_state,
                              path,
                              schema=schema,
                              created_callback=created_callback,
                              updates=updates)


class LiveDepositor(Saver, Loader):
    """Depositor with strategy that all objects that get referenced should be saved"""

    def get_snapshot_id(self, obj) -> Optional[records.SnapshotId]:
        if obj is None:
            return None

        try:
            # Try getting it from the transaction as there may be one from an in-progress save.
            # We can't use historian.get_snapshot_id here because we _only_ want one that's
            # currently being saved or we should try saving it as below to ensure it's up to date
            return self._get_current_snapshot_id(obj)
        except exceptions.NotFound:
            # Then we have to save it and get the resulting reference
            return self._historian._save_object(obj, self).snapshot_id

    def load(self, snapshot_id: records.SnapshotId):
        try:
            return self._historian.get_obj(snapshot_id.obj_id)
        except exceptions.NotFound:
            return self._historian._load_object(snapshot_id.obj_id, self)

    def load_from_record(self, record: records.DataRecord):
        """Load an object from a record"""
        with self._historian.in_transaction() as trans:

            def created(path, new_obj):
                """Called each time an object is created whilst decoding"""
                # For the root object, put it into the transaction as a live object
                if not path:
                    trans.insert_live_object(new_obj, record)

            updates = {}
            loaded = self.decode(record.state,
                                 record.get_state_schema(),
                                 created_callback=created,
                                 updates=updates)

            if updates:
                logger.warning(
                    "Object snapshot '%s' is at an older version that your current codebase.  It "
                    "can be migrated by using `mince migrate` from the command line.  If this "
                    "object is saved the new entry will use the new version.", record.snapshot_id)

            return loaded

    def update_from_record(self, obj, record: records.DataRecord) -> bool:
        """Do an in-place update of a object from a record"""
        historian = self.get_historian()
        helper = historian.get_helper(type(obj))
        with historian.in_transaction() as trans:
            # Make sure the record is in the transaction with the object
            trans.insert_live_object(obj, record)

            saved_state = self._recursive_unpack(record.state, record.get_state_schema())
            helper.load_instance_state(obj, saved_state, self)
            return True

    def save_from_builder(self, obj, builder: records.DataRecordBuilder):
        """Save a live object"""
        assert builder.snapshot_hash is not None, \
            "The snapshot hash must be set on the builder before saving"
        historian = self.get_historian()

        with historian.in_transaction() as trans:  # type: transactions.Transaction
            # Insert the object into the transaction so others can refer to it
            sid = records.SnapshotId(builder.obj_id, builder.version)
            with trans.prepare_for_saving(sid, obj):
                # trans.insert_live_object_snapshot_id(sid, obj)

                # Inject the extras
                builder.extras.update(self._get_extras(obj, builder.obj_id, builder.version))

                # Now ask the object to save itself and create the record
                builder.update(self.save_state(obj))
                record = builder.build()

                # Insert the record into the transaction
                trans.insert_live_object(obj, record)
                trans.stage(operations.Insert(record))  # Stage it for being saved

        return record

    def _get_current_snapshot_id(self, obj) -> records.SnapshotId:
        """Get the current snapshot id of an object"""
        return self._historian.current_transaction().get_snapshot_id_for_live_object(obj)

    def _get_extras(self, obj, obj_id, version: int) -> dict:
        """Get the extras dictionary for a object that is going to be saved"""
        historian = self.get_historian()
        extras = self.get_historian().get_user_info()

        if version == 0:
            # Stuff to be done the first time an object is saved
            obj_info = staging.get_info(obj)
            if obj_info:
                # Deal with a possible object creator
                created_by = obj_info.get(records.ExtraKeys.CREATED_BY, None)
                if created_by is not None:
                    try:
                        sid = historian.get_snapshot_id(created_by)
                        extras[records.ExtraKeys.CREATED_BY] = sid.obj_id
                    except exceptions.NotFound:
                        logger.info(
                            "Object with id '%s' is being saved but information about the "
                            "object it was created by will not be in the record because "
                            "the original object has not been saved yet and therefore has "
                            "no id.", obj_id)

                # Deal with possible copied from
                copied_from = obj_info.get(records.ExtraKeys.COPIED_FROM, None)
                if copied_from is not None:
                    try:
                        sid = historian.get_snapshot_id(copied_from)
                        extras[records.ExtraKeys.COPIED_FROM] = sid.to_dict()
                    except exceptions.NotFound:
                        logger.info(
                            "Object with id '%s' is being saved but information about the "
                            "object it was copied from will not be in the record because "
                            "the original object has not been saved yet and therefore has "
                            "no id.", obj_id)

        return extras


class SnapshotLoader(Loader):
    """Responsible for loading snapshots.  This object should not be reused and only
    one external call to `load` should be made.  This is because it keeps an internal
    cache."""

    def __init__(self, historian):
        super().__init__(historian)
        self._snapshots = {}  # type: MutableMapping[records.SnapshotId, Any]

    def load(self, snapshot_id: records.SnapshotId) -> Any:
        """Load an object from its snapshot id"""
        if not isinstance(snapshot_id, records.SnapshotId):
            raise TypeError(snapshot_id)

        try:
            snapshot = self._snapshots[snapshot_id]
        except KeyError:
            record = self.get_archive().load(snapshot_id)
            if record.is_deleted_record():
                snapshot = None
            else:
                snapshot = self.load_from_record(record)

            # Cache it
            self._snapshots[snapshot_id] = snapshot

        return snapshot

    def load_from_record(self, record: records.DataRecord) -> Any:
        with self._historian.in_transaction() as trans:  # type: transactions.Transaction
            updates = {}
            obj = self.decode(record.state, record.get_state_schema(), updates=updates)
            trans.insert_snapshot(obj, record.snapshot_id)
            if updates:
                logger.warning(
                    "Object snapshot '%s' is at an older version that your current codebase.  It "
                    "can be migrated by using `mince migrate` from the command line.",
                    record.snapshot_id)

            return obj


class Migrator(Saver, SnapshotLoader):
    """A migrating depositor used to make migrations to database records"""

    def get_snapshot_id(self, obj) -> records.SnapshotId:
        try:
            return self.get_historian().get_snapshot_id(obj)
        except exceptions.NotFound:
            pass

        # Ok, try the current transaction
        trans = self.get_historian().current_transaction()
        if trans is not None:
            for sid, snapshot in trans.snapshots.items():
                if obj is snapshot:
                    return sid

        # Ok, it's a brand new object that's never been saved, so save it
        self._historian.save_one(obj)
        return self.get_historian().get_snapshot_id(obj)

    def migrate_records(self,
                        to_migrate: Iterable[records.DataRecord]) -> Sequence[records.DataRecord]:
        """Migrate multiple records.  This call will return an iterable of those that were migrated
        """
        migrated = []
        with self._historian.in_transaction() as trans:  # type: transactions.Transaction
            for record in to_migrate:
                updates = {}
                obj = self.decode(record.state, record.get_state_schema(), updates=updates)
                if updates:
                    self._migrate_record(record, obj, trans)
                    migrated.append(record)

        return migrated

    def _migrate_record(self, record, new_obj, trans):
        """Given the current record and the corresponding instance this will save an updated state
        to the dictionary by re-saving the object.  The current transaction must be supplied."""
        new_schema = []
        new_state = self.encode(new_obj, new_schema)

        trans.stage(
            operations.Update(record.snapshot_id, {
                records.STATE: new_state,
                records.STATE_TYPES: new_schema
            }))

        logger.info("Snapshot %s has been migrated to the latest version", record.snapshot_id)



import pymongo.database

from mincepy import q
from . import migrate


class Initial(migrate.Migration):
    NAME = 'initial-setup'
    VERSION = 0

    def upgrade(self, database: pymongo.database.Database):
        collections = database.list_collection_names()
        if 'data' not in collections:
            database.create_collection('data')
        if 'meta' not in collections:
            database.create_collection('meta')
        super().upgrade(database)


class CollectionsSplit(migrate.Migration):
    """
    This migration makes the following changes wrt the initial database layout:

    Data collection
    ---------------

    `_id` is now the object id
    All historical and deleted records are removed.

    History collection
    ------------------

    Contains what data used to contain (including currently live objects).
    The _id field is the sref string i.e. <obj_id>#<version>

    References collection
    ---------------------

    Has been changed because of the format change of history ids.  This can
    simply be dropped as it will be lazily re-created if needed.

    Meta collection
    ---------------

    Unchanged.

    """
    NAME = 'split-data-collection'
    VERSION = 1
    PREVIOUS = Initial

    def upgrade(self, database: pymongo.database.Database):
        history_collection = 'history'
        data_collection = 'data'

        old_data = database[data_collection]
        history = database[history_collection]

        # Transform the entries to the new history format
        for entry in old_data.find():
            obj_id = entry['_id']['oid']
            version = entry['_id']['v']
            entry['_id'] = "{}#{}".format(obj_id, version)
            entry['obj_id'] = obj_id
            entry['ver'] = version
            history.insert_one(entry)

        # Ok, now rename, copy over what we need and drop
        old_data.rename('old_data')

        pipeline = self.pipeline_latest_version(history_collection)
        pipeline.append({'$match': {'state': {'$ne': '!!deleted'}}})

        new_data = database[data_collection]
        for entry in history.aggregate(pipeline):
            entry['_id'] = entry['obj_id']
            new_data.insert_one(entry)

        # Finally drop
        database.drop_collection('old_data')

        # This will be lazily re-created
        database.drop_collection('references')

        super().upgrade(database)

    @staticmethod
    def pipeline_latest_version(collection: str) -> list:
        """Returns a pipeline that will take the incoming data record documents and for each one
        find the latest version."""
        pipeline = []
        pipeline.extend([
            # Group by object id the maximum version
            {
                '$group': {
                    '_id': '$obj_id',
                    'max_ver': {
                        '$max': '$ver'
                    }
                }
            },
            # Then do a lookup against the same collection to get the records
            {
                '$lookup': {
                    'from': collection,
                    'let': {
                        'obj_id': '$_id',
                        'max_ver': '$max_ver'
                    },
                    'pipeline': [{
                        '$match': {
                            '$expr':
                                q.and_(
                                    q.eq_('$obj_id', '$$obj_id'),  # Match object id and version
                                    q.eq_('$ver', '$$max_ver')),
                        }
                    }],
                    'as': 'latest'
                }
            },
            # Now unwind and promote the 'latest' field
            {
                '$unwind': {
                    'path': '$latest'
                }
            },
            {
                '$replaceRoot': {
                    'newRoot': '$latest'
                }
            },
        ])

        return pipeline


class DropReferencesCache(migrate.Migration):
    """Here we reset the references cache (if present) as some of the code has changed"""

    NAME = 'drop-references-cache'
    VERSION = 2
    PREVIOUS = CollectionsSplit

    def upgrade(self, database: pymongo.database.Database):  # pylint: disable=no-self-use
        # This is an old one that was created by accident
        database.drop_collection("data.references")
        # And this will drop the current one
        database.drop_collection("references")


LATEST = CollectionsSplit



import uuid

import mincepy

# pylint: disable=invalid-name


class CarV0(mincepy.SimpleSavable):
    TYPE_ID = uuid.UUID('297808e4-9bc7-4f0a-9f8d-850a5f558663')
    ATTRS = ('colour', 'make')

    def __init__(self, colour: str, make: str):
        super().__init__()
        self.colour = colour
        self.make = make

    def save_instance_state(self, saver: mincepy.Saver):
        super().save_instance_state(saver)
        # Here, I decide to store as an array
        return [self.colour, self.make]

    def load_instance_state(self, saved_state, loader: mincepy.Loader):
        self.colour = saved_state[0]
        self.make = saved_state[1]


class CarV1(mincepy.SimpleSavable):
    TYPE_ID = uuid.UUID('297808e4-9bc7-4f0a-9f8d-850a5f558663')
    ATTRS = ('colour', 'make')

    class V0toV1(mincepy.ObjectMigration):
        VERSION = 1

        @classmethod
        def upgrade(cls, saved_state, loader: 'mincepy.Loader') -> dict:
            return dict(colour=saved_state[0], make=saved_state[1])

    # Set the migration
    LATEST_MIGRATION = V0toV1

    def __init__(self, colour: str, make: str):
        super().__init__()
        self.colour = colour
        self.make = make

    def save_instance_state(self, saver: mincepy.Saver):
        # I've changed my mind, I'd like to store it as a dict
        return dict(colour=self.colour, make=self.make)

    def load_instance_state(self, saved_state, loader: mincepy.Loader):
        self.colour = saved_state['colour']
        self.make = saved_state['make']


class CarV2(mincepy.SimpleSavable):
    TYPE_ID = uuid.UUID('297808e4-9bc7-4f0a-9f8d-850a5f558663')
    ATTRS = ('colour', 'make', 'reg')

    class V1toV2(mincepy.ObjectMigration):
        VERSION = 2
        PREVIOUS = CarV1.V0toV1

        @classmethod
        def upgrade(cls, saved_state, loader: 'mincepy.Loader') -> dict:
            # Augment the saved state
            saved_state['reg'] = 'unknown'
            return saved_state

    # Set the migration
    LATEST_MIGRATION = V1toV2

    def __init__(self, colour: str, make: str, reg=None):
        super().__init__()
        self.colour = colour
        self.make = make
        self.reg = reg

    def save_instance_state(self, _saver: mincepy.Saver):
        # We now add a reg field
        return dict(colour=self.colour, make=self.make, reg=self.reg)

    def load_instance_state(self, saved_state, _loader: mincepy.Loader):
        self.colour = saved_state['colour']
        self.make = saved_state['make']
        self.reg = saved_state['reg']


class HatchbackCarV0(CarV0):
    """A hatchback that inherits from CarV0"""
    TYPE_ID = uuid.UUID('d4131d3c-c140-4959-a545-21082dae9f1b')


class HatchbackCarV1(CarV1):
    """A hatchback that inherits from CarV1, simulating what would have happened if parent was
    migrated"""
    TYPE_ID = uuid.UUID('d4131d3c-c140-4959-a545-21082dae9f1b')


class StoreByValue(mincepy.SimpleSavable):
    ATTRS = ('ref',)
    TYPE_ID = uuid.UUID('40377bfc-901c-48bb-a85c-1dd692cddcae')

    def __init__(self, ref):
        super().__init__()
        self.ref = ref


class StoreByRef(mincepy.SimpleSavable):
    ATTRS = (mincepy.AsRef('ref'),)
    TYPE_ID = uuid.UUID('40377bfc-901c-48bb-a85c-1dd692cddcae')

    class ToRefMigration(mincepy.ObjectMigration):
        VERSION = 1

        @classmethod
        def upgrade(cls, saved_state, loader: 'mincepy.Loader') -> dict:
            # Replace the value stored version with a reference
            saved_state['ref'] = mincepy.ObjRef(saved_state['ref'])
            return saved_state

    # Changed my mind, want to store by value now
    LATEST_MIGRATION = ToRefMigration

    def __init__(self, ref):
        super().__init__()
        self.ref = ref


class A(mincepy.SimpleSavable):
    TYPE_ID = uuid.UUID('a50f21bc-899e-445f-baf7-0a1a373e51fc')
    ATTRS = ('migrations',)

    class Migration(mincepy.ObjectMigration):
        VERSION = 11

        @classmethod
        def upgrade(cls, saved_state, loader: 'mincepy.Loader'):
            saved_state['migrations'].append('A V11')

    LATEST_MIGRATION = Migration

    def __init__(self):
        super().__init__()
        self.migrations = []


class B(A):
    TYPE_ID = uuid.UUID('f1c07f5f-bf64-441d-8dc7-bbde65eb6fa2')

    class Migration(mincepy.ObjectMigration):
        VERSION = 2

        @classmethod
        def upgrade(cls, saved_state, loader: 'mincepy.Loader'):
            saved_state['migrations'].append('B V2')

    LATEST_MIGRATION = Migration


class BV3(A):
    TYPE_ID = uuid.UUID('f1c07f5f-bf64-441d-8dc7-bbde65eb6fa2')

    class Migration(mincepy.ObjectMigration):
        VERSION = 3
        PREVIOUS = B.Migration

        @classmethod
        def upgrade(cls, saved_state, loader: 'mincepy.Loader'):
            pass

    LATEST_MIGRATION = Migration


class C(B):
    TYPE_ID = uuid.UUID('c76153c1-82d0-4048-bdbe-937889c7fac9')


class C_BV3(BV3):
    TYPE_ID = uuid.UUID('c76153c1-82d0-4048-bdbe-937889c7fac9')



try:
    from contextlib import nullcontext
except ImportError:
    from contextlib2 import nullcontext

import pytest

import mincepy
from mincepy.testing import Car
import mincepy.testing
from . import utils


def insert_cars(historian: mincepy.Historian, num=100, in_transaction=False):
    """Insert a number of cars into the database, optionally in a transaction so they all get
    inserted in one go."""
    if in_transaction:
        ctx = historian.transaction()
    else:
        ctx = nullcontext()

    with ctx:
        for _ in range(num):
            historian.save(mincepy.testing.Car())


def find(historian, **kwargs):
    # Have to wrap the find method like this because it returns a generator and won't necessarily
    # fetch from the db unless we iterate it
    return tuple(historian.find(**kwargs))


def test_benchmark_insertions_individual(historian: mincepy.Historian, benchmark):
    benchmark(insert_cars, historian, in_transaction=False)


def test_benchmark_insertions_transaction(historian: mincepy.Historian, benchmark):
    benchmark(insert_cars, historian, in_transaction=True)


@pytest.mark.parametrize("num", [10**i for i in range(5)])
def test_find_cars(historian: mincepy.Historian, benchmark, num):
    """Test finding a car as a function of the number of entries in the database"""
    # Put in the one we want to find
    historian.save(Car('honda', 'green'))

    # Put in the correct number of random other entries
    for _ in range(num):
        historian.save(Car(utils.random_str(10), utils.random_str(5)))

    result = benchmark(find, historian, state=dict(make='honda', colour='green'))
    assert len(result) == 1


@pytest.mark.parametrize("num", [5**i for i in range(1, 4)])
def test_load_cars(historian: mincepy.Historian, benchmark, num):
    """Test finding a car as a function of the number of entries in the database"""
    # Put in the correct number of random other entries
    car_ids = []
    for _ in range(num):
        car_ids.append(historian.save(Car(utils.random_str(10), utils.random_str(5))))

    result = benchmark(historian.load, *car_ids)
    assert len(result) == len(car_ids)



"""Module containing functions to generate query operations.  To prevent clashes with python
builtins we append underscores to the function names.  This also makes it safer to import this
module as a wildcard import.
"""

__all__ = ('and_', 'or_', 'eq_', 'exists_', 'elem_match_', 'gt_', 'gte_', 'in_', 'lt_', 'lte_',
           'ne_', 'nin_')

# region Logical operators


def and_(*conditions) -> dict:
    """Helper that produces query dict for AND of multiple conditions"""
    if len(conditions) == 1:
        return conditions[0]

    return {'$and': list(conditions)}


def or_(*conditions) -> dict:
    """Helper that produces query dict for OR of multiple conditions"""
    if len(conditions) == 1:
        return conditions[0]

    return {'$or': list(conditions)}


# endregion


def eq_(one, other) -> dict:
    """Helper that produces mongo query dict for two items being equal"""
    return {'$eq': [one, other]}


# region Element operators


def exists_(key, value: bool = True) -> dict:
    """Return condition for the existence of a key.  If True, matches if key exists, if False
    matches if it does not."""
    return {key: {'$exists': value}}


# endregion

# region Array operators


def elem_match_(**conditions) -> dict:
    """Match an element that is an array and has at least one member that matches all the specified
    conditions"""
    return {'$elemMatch': conditions}


# endregion

# region Comparison operators


def gt_(quantity) -> dict:
    """Match values greater than quantity"""
    return {'$gt': quantity}


def gte_(quantity) -> dict:
    """Match values greater than or equal to quantity"""
    return {'$gte': quantity}


def in_(*possibilities) -> dict:
    """Match values that are equal to any of the possibilities"""
    if len(possibilities) == 1:
        return possibilities[0]

    return {'$in': list(possibilities)}


def lt_(quantity) -> dict:
    """Match values less than quantity"""
    return {'$lt': quantity}


def lte_(quantity) -> dict:
    """Match values less than or equal to quantity"""
    return {'$lte': quantity}


def ne_(value) -> dict:
    """Match values not equal to to value"""
    return {'$ne': value}


def nin_(*possibilities) -> dict:
    """Match values that are not equal to any of the possibilities"""
    if len(possibilities) == 1:
        return possibilities[0]

    return {'$nin': list(possibilities)}


# endregion



""""Tests of migrations"""
import gc
import logging
import uuid

import pytest

import mincepy
from mincepy import testing
from .common import CarV0, CarV1, CarV2, StoreByValue, StoreByRef

# pylint: disable=invalid-name


def test_simple_migration(historian: mincepy.Historian):
    car = CarV0('red', 'ferrari')
    car_id = historian.save(car)
    del car

    # Now change to version 2
    historian.register_type(CarV1)
    loaded_car = historian.load(car_id)
    assert loaded_car.colour == 'red'
    assert loaded_car.make == 'ferrari'


def test_multiple_migrations(historian: mincepy.Historian):
    car = CarV0('red', 'ferrari')
    car_id = historian.save(car)
    del car

    # Now change to version 3, skipping 2
    historian.register_type(CarV2)
    loaded_car = historian.load(car_id)
    assert loaded_car.colour == 'red'
    assert loaded_car.make == 'ferrari'
    assert hasattr(loaded_car, 'reg')
    assert loaded_car.reg == 'unknown'


def test_migrate_to_reference(historian: mincepy.Historian, caplog):
    caplog.set_level(logging.DEBUG)

    car = testing.Car()
    by_val = StoreByValue(car)
    oid = historian.save(by_val)
    del by_val

    # Now, change my mind
    historian.register_type(StoreByRef)
    # Migrate
    migrated = historian.migrations.migrate_all()
    assert len(migrated) == 1

    by_ref = historian.load(oid)
    assert isinstance(by_ref.ref, testing.Car)
    assert by_ref.ref is not car
    loaded_car = by_ref.ref
    del by_ref

    # Reload and check that ref points to the same car
    loaded = historian.load(oid)
    assert loaded.ref is loaded_car


def test_dependent_migrations(historian: mincepy.Historian, caplog):
    """Test what happens when both a parent object and a contained object need migration"""
    caplog.set_level(logging.DEBUG)

    car = CarV0('red', 'honda')
    by_val = StoreByValue(car)

    oid = historian.save(by_val)
    del by_val

    loaded = historian.load(oid)
    assert isinstance(loaded.ref, CarV0)
    assert loaded.ref is not car
    del loaded

    # Now, change my mind
    historian.register_type(StoreByRef)  # Store by reference instead
    historian.register_type(CarV2)  # And update the car
    # Migrate
    migrated = historian.migrations.migrate_all()
    assert len(migrated) == 1  # Just by_val

    by_ref = historian.load(oid)
    assert isinstance(by_ref.ref, CarV2)
    loaded_car = by_ref.ref
    assert by_ref is not car

    # Now delete, load and check we have the same car
    del by_ref
    reloaded = historian.load(oid)
    assert reloaded.ref is loaded_car


def test_migrating_snapshot(historian: mincepy.Historian, caplog):
    """Test migrating an out of date snapshot"""
    caplog.set_level(logging.INFO)

    car = CarV0('yellow', 'bugatti')
    car_id = car.save()  # Version 0

    car.colour = 'brown'
    car.save()  # Version 1
    del car

    historian.register_type(CarV2)  # And update the car definition
    snapshot = historian.load_snapshot(mincepy.SnapshotId(car_id, 0))  # Load version 0

    assert snapshot.colour == 'yellow'
    assert snapshot.make == 'bugatti'

    # Now load the current version
    current = historian.load(car_id)
    assert current.colour == 'brown'
    assert current.make == 'bugatti'


def test_migrating_live_object(historian: mincepy.Historian):
    """Test that a migration including a live object works fine"""

    class V1(mincepy.SimpleSavable):
        TYPE_ID = uuid.UUID('8b1620f6-dd6d-4d39-b8b1-4433dc2a54df')
        ATTRS = ('ref',)

        def __init__(self, obj):
            super().__init__()
            self.ref = obj

    car = testing.Car()
    car.save()

    class V2(mincepy.SimpleSavable):
        TYPE_ID = uuid.UUID('8b1620f6-dd6d-4d39-b8b1-4433dc2a54df')
        ATTRS = (mincepy.AsRef('ref'),)

        class V1toV2(mincepy.ObjectMigration):
            VERSION = 1

            @classmethod
            def upgrade(cls, saved_state, loader: 'mincepy.Loader'):
                # Create a reference to the live car object
                saved_state['ref'] = mincepy.ObjRef(car)
                return saved_state

        LATEST_MIGRATION = V1toV2

    martin = testing.Person('martin', 35)
    my_obj = V1(martin)
    my_obj_id = my_obj.save()
    del my_obj

    # Now change my mind
    historian.register_type(V2)
    assert len(historian.migrations.migrate_all()) == 1

    migrated = historian.load(my_obj_id)
    assert migrated.ref is car


def test_loading_newer_version_no_migrations(historian: mincepy.Historian):
    """Test loading an object that has a newer version when we have no migrations what so ever"""
    car = CarV1('blue', 'honda')
    car_id = car.save()
    del car

    # Now, pretend we're a user with an older codebase
    historian.register_type(CarV0)

    with pytest.raises(mincepy.VersionError):
        historian.load(car_id)

    # This shouldn't raise as there are no migrations to be done from _this_ codebase
    historian.migrations.migrate_all()


def test_loading_newer_version_older_migrations(historian: mincepy.Historian):
    """Test loading an object that has a newer version than our latest migration"""
    car = CarV2('blue', 'honda')
    car_id = car.save()
    del car

    # Now, pretend we're a user with an older codebase
    historian.register_type(CarV1)

    with pytest.raises(mincepy.VersionError):
        historian.load(car_id)

    historian.migrations.migrate_all()


def test_loading_nested_migrations(historian: mincepy.Historian):
    """Test migrating an object that itself contains an instance of its own type, both need
    migration"""
    car = testing.Car()
    obj = StoreByValue(car)
    parent = StoreByValue(obj)
    obj_id = parent.save()
    del parent, obj, car

    historian.register_type(StoreByRef)
    parent = historian.load(obj_id)
    assert isinstance(parent.ref, StoreByRef)
    assert isinstance(parent.ref.ref, testing.Car)


def test_lazy_migrating_with_saved(historian: mincepy.Historian):
    """Test migrating an object that has saved references"""

    class V3(mincepy.SimpleSavable):
        ATTRS = mincepy.AsRef('ref'), 'description'
        TYPE_ID = uuid.UUID('40377bfc-901c-48bb-a85c-1dd692cddcae')

        class Migration(mincepy.ObjectMigration):
            VERSION = 2
            PREVIOUS = StoreByRef.ToRefMigration

            @classmethod
            def upgrade(cls, saved_state, loader: 'mincepy.Loader'):
                saved_state['description'] = None

        LATEST_MIGRATION = Migration

        def __init__(self, ref):
            super().__init__()
            self.ref = ref
            self.description = None

    obj = StoreByRef(testing.Car())
    obj_id = obj.save()
    del obj
    gc.collect()

    historian.register_type(V3)
    obj = historian.load(obj_id)

    assert isinstance(obj, V3)
    assert hasattr(obj, 'description')
    assert obj.description is None



import functools

import pymongo

import mincepy
from mincepy import q
from . import db

# To present 'op' being flagged
# pylint: disable=invalid-name


@functools.singledispatch
def to_mongo_op(op: mincepy.operations.Operation):
    """Convert a mincepy operation to a mongodb one.  Returns a tuple of the data operation and
    history operation that need to be bulk written"""
    raise NotImplementedError


@to_mongo_op.register(mincepy.operations.Insert)
def _(op: mincepy.operations.Insert):
    """Insert"""
    record = op.record
    document = db.to_document(record)
    document['_id'] = record.obj_id

    if record.is_deleted_record():
        data_op = pymongo.operations.DeleteOne(filter={
            db.OBJ_ID: record.obj_id,
            db.VERSION: q.lt_(record.version)
        },)
    else:
        data_op = pymongo.operations.ReplaceOne(filter={
            db.OBJ_ID: record.obj_id,
            db.VERSION: q.lt_(record.version)
        },
                                                replacement=document.copy(),
                                                upsert=True)

    # History uses the sid as the document id
    document['_id'] = str(record.snapshot_id)
    history_op = pymongo.operations.InsertOne(document)

    return data_op, history_op


@to_mongo_op.register(mincepy.operations.Update)
def _(op: mincepy.operations.Update):
    """Update"""
    sid = op.snapshot_id
    update = db.to_document(op.update)
    update = {'$set': update}

    # It's fine if either (or both) of these fail to find anything to update
    data_op = pymongo.operations.UpdateOne(filter={
        db.OBJ_ID: sid.obj_id,
        db.VERSION: sid.version
    },
                                           update=update)
    history_op = pymongo.operations.UpdateOne(filter={'_id': str(sid)}, update=update)

    return data_op, history_op


@to_mongo_op.register(mincepy.operations.Delete)
def _(op: mincepy.operations.Delete):
    """Delete"""
    sid = op.snapshot_id

    # It's fine if either (or both) of these fail to find anything to update
    data_op = pymongo.operations.DeleteOne(filter={db.OBJ_ID: sid.obj_id},)
    history_op = pymongo.operations.DeleteOne(filter={'_id': str(sid)},)

    return data_op, history_op



import abc
from typing import Generic, TypeVar, NamedTuple, Sequence, Union, Mapping, Iterable, Dict, \
    Iterator, Any, Type, Optional

import networkx

from . import qops as q
from . import records
from .records import DataRecord
from . import operations

__all__ = 'Archive', 'BaseArchive', 'ASCENDING', 'DESCENDING', 'OUTGOING', 'INCOMING'

IdT = TypeVar('IdT')  # The archive ID type

# Sort options
ASCENDING = 1
DESCENDING = -1

OUTGOING = 1
INCOMING = -1


class Archive(Generic[IdT], metaclass=abc.ABCMeta):
    """An archive provides the persistent storage for the historian.  It is responsible for storing,
    searching and loading data records and their metadata."""

    # pylint: disable=too-many-public-methods

    SnapshotId = records.SnapshotId[IdT]
    MetaEntry = NamedTuple('MetaEntry', [('obj_id', IdT), ['meta', dict]])

    @classmethod
    def get_types(cls) -> Sequence:
        """This method allows the archive to return either types or type helper that the historian
        should support.  A common example is the type helper for the object id type"""
        return tuple()

    @classmethod
    @abc.abstractmethod
    def get_id_type(cls) -> Type[IdT]:
        """Get the type used as an ID by this archive"""

    @abc.abstractmethod
    def create_archive_id(self) -> IdT:
        """Create a new archive id"""

    @abc.abstractmethod
    def construct_archive_id(self, value) -> IdT:
        """If it's possible, construct an archive value from the passed value.
        This is useful as a convenience to the user if, say, the archive id can be
        constructed from a string.  Raise TypeError or ValueError if this is not possible
        for the given value.
        """

    @abc.abstractmethod
    def create_file(self, filename: str = None, encoding: str = None):
        """Create a new file object specific for this archive type"""

    @abc.abstractmethod
    def save(self, record: DataRecord):
        """Save a data record to the archive"""

    @abc.abstractmethod
    def save_many(self, data_records: Sequence[DataRecord]):
        """Save many data records to the archive"""

    @abc.abstractmethod
    def bulk_write(self, ops: Sequence[operations.Operation]):
        """Made a collection of write operations to the database"""

    # region Metadata

    @abc.abstractmethod
    def meta_get(self, obj_id: IdT) -> Optional[dict]:
        """Get the metadata for an objects."""

    @abc.abstractmethod
    def meta_get_many(self, obj_ids: Iterable[IdT]) -> Dict:
        """Get the metadata for multiple objects.  Returns a dictionary mapping the object id to
        the metadata dictionary"""

    @abc.abstractmethod
    def meta_set(self, obj_id: IdT, meta: Optional[Mapping]):
        """Set the metadata on on the object with the corresponding id"""

    @abc.abstractmethod
    def meta_set_many(self, metas: Mapping[IdT, Optional[Mapping]]):
        """Set the metadata on multiple objects.  This takes a mapping of the object id to the
        corresponding (optional) metadata dictionary"""

    @abc.abstractmethod
    def meta_update(self, obj_id: IdT, meta: Mapping):
        """Update the metadata on the object with the corresponding id"""

    @abc.abstractmethod
    def meta_update_many(self, metas: Mapping[IdT, Mapping]):
        """Update the metadata on multiple objects.  This method expects to get a mapping of object
        id to the mapping to be used to update the metadata for that object"""

    @abc.abstractmethod
    def meta_find(
            self,
            filter: dict = None,  # pylint: disable=redefined-builtin
            obj_id: Union[IdT, Iterable[IdT], Mapping] = None) -> \
            'Iterator[Archive.MetaEntry]':
        """Yield metadata satisfying the given criteria.  The search can optionally be restricted to
        a set of passed object ids.

        :param filter: a query filter for the search
        :param obj_id: an optional restriction on the object ids to search.  This ben be either:
            1. a single object id
            2. an iterable of object ids in which is treated as {'$in': list(obj_ids)}
            3. a general query filter to be applied to the object ids
        """

    @abc.abstractmethod
    def meta_distinct(
            self,
            key: str,
            filter: dict = None,  # pylint: disable=redefined-builtin
            obj_id: Union[IdT, Iterable[IdT], Mapping] = None) -> \
            'Iterator':
        """Yield distinct values found for 'key' within metadata documents, optionally marching a
        search filter.

        The search can optionally be restricted to a set of passed object ids.

        :param key: the document key to get distinct values for
        :param filter: a query filter for the search
        :param obj_id: an optional restriction on the object ids to search.  This ben be either:
            1. a single object id
            2. an iterable of object ids in which is treated as {'$in': list(obj_ids)}
            3. a general query filter to be applied to the object ids
        """

    @abc.abstractmethod
    def meta_create_index(self, keys, unique=False, where_exist=False):
        """Create an index on the metadata.  Takes either a single key or list of (key, direction)
         pairs

         :param keys: the key or keys to create the index on
         :param unique: if True, create a uniqueness constraint on this index
         :param where_exist: if True the index only applies for documents where the key(s) exist
         """

    # endregion

    @abc.abstractmethod
    def load(self, snapshot_id: SnapshotId) -> DataRecord:
        """Load a snapshot of an object with the given reference"""

    @abc.abstractmethod
    def history(self, obj_id: IdT, idx_or_slice) -> [DataRecord, Sequence[DataRecord]]:
        """Load the snapshot records for a particular object, can return a single or multiple
        records"""

    @abc.abstractmethod
    def get_snapshot_ids(self, obj_id: IdT) -> 'Sequence[Archive.SnapshotId]':
        """Returns a list of time ordered snapshot ids"""

    # pylint: disable=too-many-arguments
    @abc.abstractmethod
    def find(self,
             obj_id: Union[IdT, Iterable[IdT], Dict] = None,
             type_id=None,
             created_by=None,
             copied_from=None,
             version=None,
             state=None,
             state_types=None,
             snapshot_hash=None,
             meta: dict = None,
             extras: dict = None,
             limit=0,
             sort=None,
             skip=0) -> Iterator[DataRecord]:
        """Find records matching the given criteria

        :param type_id: the type id to look for
        :param created_by: find records with the given type id
        :param copied_from: find records copied from the record with the given id
        :param version: restrict the search to this version, -1 for latest
        :param state: find objects with this state filter
        :param state_types: file objects with this state types filter
        :param snapshot_hash: find objects with this snapshot hash
        :param meta: find objects with this meta filter
        :param extras: the search criteria to apply on the data record extras
        :param limit: limit the results to this many records
        :param obj_id: an optional restriction on the object ids to search.  This ben be either:
            1. a single object id
            2. an iterable of object ids in which is treated as {'$in': list(obj_ids)}
            3. a general query filter to be applied to the object ids
        :param sort: sort the results by the given criteria
        :param skip: skip the this many entries
        """

    @abc.abstractmethod
    def count(self,
              obj_id=None,
              type_id=None,
              created_by=None,
              copied_from=None,
              version=-1,
              state=None,
              snapshot_hash=None,
              meta=None,
              limit=0):
        """Count the number of entries that match the given query"""

    @abc.abstractmethod
    def get_snapshot_ref_graph(self,
                               *snapshot_ids: SnapshotId,
                               direction=OUTGOING,
                               max_dist: int = None) -> networkx.DiGraph:
        """Given one or more snapshot ids the archive will supply the corresponding reference
        graph(s).  The graphs start at the given id and contains all snapshots that it references,
        all snapshots they reference and so on.
        """

    @abc.abstractmethod
    def get_obj_ref_graph(self,
                          *obj_ids: IdT,
                          direction=OUTGOING,
                          max_dist: int = None) -> networkx.DiGraph:
        """Given one or more object ids the archive will supply the corresponding reference
        graph(s).  The graphs start at the given id and contains all object ids that it references,
        all object ids they reference and so on.
        """


class BaseArchive(Archive[IdT]):
    ID_TYPE = None  # type: Type[IdT]

    @classmethod
    def get_id_type(cls) -> Type[IdT]:
        assert cls.ID_TYPE, "The ID type has not been set on this archive"
        return cls.ID_TYPE

    def save(self, record: DataRecord):
        return self.bulk_write([operations.Insert(record)])

    def save_many(self, data_records: Sequence[DataRecord]):
        """
        This will save records one by one but subclass may want to override this behaviour if
        they can save multiple records at once.
        """
        self.bulk_write([operations.Insert(record) for record in data_records])

    def meta_get_many(self, obj_ids: Iterable[IdT]) -> Dict[IdT, dict]:
        metas = {}
        for obj_id in obj_ids:
            metas[obj_id] = self.meta_get(obj_id)
        return metas

    def meta_update_many(self, metas: Mapping[IdT, Mapping]):
        for entry in metas.items():
            self.meta_update(*entry)

    def meta_set_many(self, metas: Mapping[IdT, Mapping]):
        for entry in metas.items():
            self.meta_set(*entry)

    def history(self, obj_id: IdT, idx_or_slice) -> [DataRecord, Sequence[DataRecord]]:
        refs = self.get_snapshot_ids(obj_id)[idx_or_slice]
        if len(refs) > 1:
            return [self.load(ref) for ref in refs]

        # Single one
        return self.load(refs[0])

    def construct_archive_id(self, value) -> IdT:  # pylint: disable=no-self-use
        raise TypeError("Not possible to construct an archive id from '{}'".format(type(value)))


def scalar_query_spec(specifier: Union[Mapping, Iterable[Any], Any]) -> \
        Union[Any, Dict]:
    """Convenience function to create a query specifier for a given item.  There are three
    possibilities:

    1. The item is a mapping in which case it is returned as is.
    2. The item is an iterable (but not a mapping) in which case it is interpreted to mean:
        {'$in': list(iterable)}
    3. it is a raw item item in which case it is matched directly
    """
    if isinstance(specifier, dict):  # This has to be first as dict is iterable
        return specifier
    if isinstance(specifier, Iterable):
        return q.in_(*specifier)

    return specifier



from . import common_helpers
from . import builtins
from . import refs


def get_types():
    """Provide a list of all historian types"""
    types = list()
    types.extend(builtins.HISTORIAN_TYPES)
    types.extend(common_helpers.HISTORIAN_TYPES)
    types.extend(refs.HISTORIAN_TYPES)

    return types



import logging
import sys

import click

from . import main


@main.mince.group()
def dev():
    """Commands to help with development and testing"""


@dev.command()
@click.argument('uri', type=str)
def populate(uri):
    """Populate the database with testing data"""
    import mincepy.testing  # pylint: disable=import-outside-toplevel
    main.set_print_logging(logging.INFO)
    try:
        hist = mincepy.create_historian(uri)
    except ValueError as exc:
        click.echo(exc)
        sys.exit(1)
    else:
        mincepy.testing.populate(hist)
        click.echo("Successfully populated database at '{}'".format(uri))



"""Module for all the built in container and other types"""

from abc import ABCMeta, abstractmethod
import collections
from pathlib import Path
import shutil
from typing import BinaryIO, Optional
import uuid

from . import base_savable
from . import helpers
from . import records
from . import refs
from .utils import sync
from . import types

__all__ = ('List', 'LiveList', 'LiveRefList', 'RefList', 'Str', 'Dict', 'RefDict', 'LiveDict',
           'LiveRefDict', 'BaseFile', 'File')


class _UserType(base_savable.SimpleSavable, metaclass=ABCMeta):
    """Mixin for helping user types to be compatible with the historian.
    These typically have a .data member that stores the actual data (list, dict, str, etc)"""
    ATTRS = ('data',)
    data = None  # placeholder
    # This is an optional type for the data member.  See save_instance_state type for information
    # on when it might be useful
    DATA_TYPE = None  # type: type

    def save_instance_state(self, saver):
        # This is a convenient way of storing primitive data types directly as the state
        # rather than having to be a 'data' member of a dictionary.  This makes it much
        # easier to search these types
        if self.DATA_TYPE is not None and \
                issubclass(self.DATA_TYPE, saver.get_historian().primitives):
            self._on_save(saver)  # Call this here are we aren't going to call up the hierarchy
            return self.data

        return super().save_instance_state(saver)

    def load_instance_state(self, saved_state, loader):
        # See save_instance_state
        if self.DATA_TYPE is not None and \
                issubclass(self.DATA_TYPE, loader.get_historian().primitives):
            self._on_load(loader)  # Call this here are we aren't going to call up the hierarchy
            self.data = saved_state
        else:
            super(_UserType, self).load_instance_state(saved_state, loader)


class ObjProxy(_UserType):
    """A simple proxy for any object/data type which can also be a primitive"""
    TYPE_ID = uuid.UUID('d43c2db5-1e8c-428f-988f-8b198accde47')

    def __init__(self, data=None):
        super(ObjProxy, self).__init__()
        self.data = data

    def __call__(self):
        return self.data

    def assign(self, value):
        self.data = value


class Str(collections.UserString, _UserType):
    TYPE_ID = uuid.UUID('350f3634-4a6f-4d35-b229-71238ce9727d')
    DATA_TYPE = str

    def __init__(self, seq):
        collections.UserString.__init__(self, seq)
        _UserType.__init__(self)


class Reffer:
    """Mixin for types that want to be able to create references non-primitive types"""

    # pylint: disable=too-few-public-methods, no-self-use

    def _ref(self, obj):
        """Create a reference for a non-primitive, otherwise use the value"""
        if types.is_primitive(obj):
            return obj

        return self._create_ref(obj)

    def _deref(self, obj):
        """Dereference a non-primitive, otherwise return the value"""
        if types.is_primitive(obj):
            return obj

        return obj()

    def _create_ref(self, obj):
        """Create a reference for a given object"""
        return refs.ObjRef(obj)


# region lists


class List(collections.UserList, _UserType):
    TYPE_ID = uuid.UUID('2b033f70-168f-4412-99ea-d1f131e3a25a')
    DATA_TYPE = list

    def __init__(self, initlist=None):
        collections.UserList.__init__(self, initlist)
        _UserType.__init__(self)


class RefList(collections.abc.MutableSequence, Reffer, _UserType):
    """A list that stores all entries as references in the database except primitives"""
    TYPE_ID = uuid.UUID('091efff5-136d-4ac2-bd59-28f50f151263')
    DATA_TYPE = list

    def __init__(self, init_list=None):
        super().__init__()
        init_list = init_list or []
        self.data = self.DATA_TYPE(self._ref(item) for item in init_list)

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return repr(self.data)

    def __getitem__(self, item):
        return self._deref(self.data[item])

    def __setitem__(self, key, value):
        self.data[key] = self._ref(value)

    def __delitem__(self, key):
        self.data.__delitem__(key)

    def __len__(self):
        return self.data.__len__()

    def insert(self, index, value):
        self.data.insert(index, self._ref(value))


class LiveList(collections.abc.MutableSequence, _UserType):
    """A list that is always in sync with the database"""
    TYPE_ID = uuid.UUID('c83e6206-cd29-4fda-bf76-11fce1681cd9')

    def __init__(self, init_list=None):
        super(LiveList, self).__init__()
        init_list = init_list or []
        self.data = RefList(self._create_proxy(item) for item in init_list)

    @sync()
    def __getitem__(self, item):
        proxy = self.data[item]  # type: ObjProxy
        if self.is_saved():
            proxy.sync()
        return proxy()

    @sync(save=True)
    def __setitem__(self, key, value):
        proxy = self.data[key]  # type: ObjProxy
        proxy.assign(value)

    @sync(save=True)
    def __delitem__(self, key):
        proxy = self.data[key]  # type: ObjProxy
        del self.data[key]
        if self._historian is not None:
            self._historian.delete(proxy)

    @sync()
    def __len__(self):
        return len(self.data)

    @sync(save=True)
    def insert(self, index, value):
        proxy = self._create_proxy(value)
        self.data.insert(index, proxy)

    @sync(save=True)
    def append(self, value):
        proxy = self._create_proxy(value)
        self.data.append(proxy)

    def _create_proxy(self, obj):  # pylint: disable=no-self-use
        return ObjProxy(obj)


class LiveRefList(Reffer, LiveList):
    """A live list that uses references to store objects"""
    TYPE_ID = uuid.UUID('98454806-c587-4fcc-a514-65fdefb0180d')

    @sync()
    def __getitem__(self, item):
        proxy = self.data[item]  # type: ObjProxy
        if self.is_saved():
            proxy.sync()
        ref = proxy()
        return self._deref(ref)

    @sync(save=True)
    def __setitem__(self, key, value):
        proxy = self.data[key]  # type: ObjProxy
        proxy.assign(self._ref(value))

    def _create_proxy(self, obj):
        return ObjProxy(self._ref(obj))


# endregion

# region Dicts


class Dict(collections.UserDict, _UserType):
    TYPE_ID = uuid.UUID('a7584078-95b6-4e00-bb8a-b077852ca510')
    DATA_TYPE = dict

    def __init__(self, *args, **kwarg):
        collections.UserDict.__init__(self, *args, **kwarg)
        _UserType.__init__(self)


class RefDict(collections.MutableMapping, Reffer, _UserType):
    """A dictionary that stores all values as references in the database."""
    TYPE_ID = uuid.UUID('c95f4c4e-766b-4dda-a43c-5fca4fd7bdd0')
    DATA_TYPE = dict

    def __init__(self, *args, **kwargs):
        super().__init__()
        initial = dict(*args, **kwargs)
        self.data = self.DATA_TYPE({key: self._ref(value) for key, value in initial.items()})

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return repr(self.data)

    def __getitem__(self, item):
        return self._deref(self.data[item])

    def __setitem__(self, key, value):
        self.data[key] = self._ref(value)

    def __delitem__(self, key):
        self.data.__delitem__(key)

    def __iter__(self):
        return self.data.__iter__()

    def __len__(self):
        return self.data.__len__()


class LiveDict(collections.MutableMapping, _UserType):
    TYPE_ID = uuid.UUID('740cc832-721c-4f85-9628-706257eb55b9')
    DATA_TYPE = RefDict

    def __init__(self, *args, **kwargs):
        super().__init__()
        initial = dict(*args, **kwargs)
        self.data = RefDict({key: self._create_proxy(value) for key, value in initial.items()})

    @sync()
    def __getitem__(self, item):
        proxy = self.data[item]  # type: ObjProxy
        proxy.sync()
        return proxy()

    @sync()
    def __setitem__(self, key, value):
        if key in self.data:
            # Can simply update the proxy
            proxy = self.data[key]  # type: ObjProxy
            proxy.assign(value)
            if proxy.is_saved():
                proxy.save()
        else:
            proxy = self._create_proxy(value)
            self.data[key] = proxy
            if self.is_saved():
                self.save()  # This will cause the proxy to be saved as well

    @sync(save=True)
    def __delitem__(self, key):
        proxy = self.data[key]
        del self.data[key]
        self._historian.delete(proxy)

    @sync()
    def __iter__(self):
        return self.data.__iter__()

    @sync()
    def __len__(self):
        return len(self.data)

    def _create_proxy(self, value):  # pylint: disable=no-self-use
        return ObjProxy(value)


class LiveRefDict(Reffer, LiveDict):
    """A live dictionary that uses references to refer to contained objects"""
    TYPE_ID = uuid.UUID('16e7e814-8268-46e0-8d8e-6f34132366b9')

    def __init__(self, *args, **kwargs):
        super().__init__()

        initial = dict(*args, **kwargs)
        self.data = RefDict({key: self._create_proxy(value) for key, value in initial.items()})

    @sync()
    def __getitem__(self, item):
        proxy = self.data[item]  # type: ObjProxy
        proxy.sync()
        ref = proxy()
        return self._deref(ref)

    @sync()
    def __setitem__(self, key, value):
        if key in self.data:
            # Can simply update the proxy
            proxy = self.data[key]  # type: ObjProxy
            proxy.assign(self._ref(value))
            if proxy.is_saved():
                proxy.save()
        else:
            proxy = self._create_proxy(value)
            self.data[key] = proxy
            if self.is_saved():
                self.save()  # This will cause the proxy to be saved as well

    def _create_proxy(self, value):
        return ObjProxy(self._ref(value))


# endregion


class File(base_savable.SimpleSavable, metaclass=ABCMeta):
    ATTRS = ('_filename', '_encoding')
    READ_SIZE = 256  # The number of bytes to read at a time

    def __init__(self, filename: str = None, encoding=None):
        super().__init__()
        self._filename = filename
        self._encoding = encoding

    @property
    def filename(self) -> Optional[str]:
        return self._filename

    @property
    def encoding(self) -> Optional[str]:
        return self._encoding

    @abstractmethod
    def open(self, mode='r', **kwargs) -> BinaryIO:
        """Open returning a file like object that supports close() and read()"""

    def from_disk(self, path):
        """Copy the contents of a disk file to this file"""
        with open(str(path), 'r', encoding=self.encoding) as disk_file:
            with self.open('w') as this:
                shutil.copyfileobj(disk_file, this)

    def to_disk(self, folder: [str, Path]):
        """Copy the contents of this file to a file on disk in the given folder"""
        file_path = Path(str(folder)) / self.filename
        with open(str(file_path), 'w', encoding=self._encoding) as disk_file:
            with self.open('r') as this:
                shutil.copyfileobj(this, disk_file)

    def write_text(self, text: str, encoding=None):
        encoding = encoding or self._encoding
        with self.open('w', encoding=encoding) as fileobj:
            fileobj.write(text)

    def read_text(self, encoding=None) -> str:
        """Read the contents of the file as text.
        This function is named as to mirror pathlib.Path"""
        encoding = encoding or self._encoding
        with self.open('r', encoding=encoding) as fileobj:
            return fileobj.read()

    def __str__(self):
        contents = [str(self._filename)]
        if self._encoding is not None:
            contents.append("({})".format(self._encoding))
        return " ".join(contents)

    def __eq__(self, other) -> bool:
        """Compare the contents of two files

        If both files do not exist they are considered equal.
        """
        if not isinstance(other, BaseFile) or self.filename != other.filename:
            return False

        try:
            with self.open() as my_file:
                try:
                    with other.open() as other_file:
                        while True:
                            my_line = my_file.readline(self.READ_SIZE)
                            other_line = other_file.readline(self.READ_SIZE)
                            if my_line != other_line:
                                return False
                            if my_line == '' and other_line == '':
                                return True
                except FileNotFoundError:
                    return False
        except FileNotFoundError:
            # Our file doesn't exist, make sure the other doesn't either
            try:
                with other.open():
                    return False
            except FileNotFoundError:
                return True

    def yield_hashables(self, hasher):
        """Hash the contents of the file"""
        try:
            with self.open('rb') as opened:
                while True:
                    line = opened.read(self.READ_SIZE)
                    if line == b'':
                        return
                    yield line
        except FileNotFoundError:
            yield from hasher.yield_hashables(None)


BaseFile = File


class SnapshotIdHelper(helpers.TypeHelper):
    """Add ability to store references"""
    TYPE = records.SnapshotId
    TYPE_ID = uuid.UUID('05fe092b-07b3-4ffc-8cf2-cee27aa37e81')

    def eq(self, one, other):  # pylint: disable=invalid-name, no-self-use
        if not (isinstance(one, records.SnapshotId) and isinstance(other, records.SnapshotId)):
            return False

        return one.obj_id == other.obj_id and one.version == other.version

    def yield_hashables(self, obj, hasher):  # pylint: disable=no-self-use
        yield from hasher.yield_hashables(obj.obj_id)
        yield from hasher.yield_hashables(obj.version)

    def save_instance_state(self, obj, _saver):  # pylint: disable=no-self-use
        return obj.to_dict()

    def load_instance_state(self, obj, saved_state, _loader):  # pylint: disable=no-self-use
        if isinstance(saved_state, list):
            # Legacy version
            obj.__init__(*saved_state)
        else:
            # New version is a dictionary
            obj.__init__(**saved_state)


HISTORIAN_TYPES = (Str, List, RefList, LiveList, LiveRefList, Dict, RefDict, LiveDict, LiveRefDict,
                   ObjProxy)



from abc import ABCMeta, abstractmethod
import logging
from typing import Type, Optional, Sequence

import pytray.pretty

import mincepy  # pylint: disable=unused-import
from . import exceptions
from . import migrations
from . import tracking
from . import types

__all__ = 'TypeHelper', 'WrapperHelper', 'BaseHelper'

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def inject_creation_tracking(cls: Type):
    # Check to make sure we don't do this twice!
    if not hasattr(cls, '__orig_new'):
        cls.__orig_new = cls.__new__  # pylint: disable=protected-access

        def new(_cls, *_args, **_kwargs):
            inst = cls.__orig_new(_cls)  # pylint: disable=protected-access
            tracking.obj_created(inst)
            return inst

        cls.__new__ = new


def remove_creation_tracking(cls: Type):
    try:
        cls.__new__ = cls.__orig_new  # pylint: disable=protected-access
    except AttributeError:
        pass


class TypeHelper(metaclass=ABCMeta):
    """This interface provides the basic methods necessary to enable a type to be compatible with
    the historian."""
    TYPE = None  # The type this helper corresponds to
    TYPE_ID = None  # The unique id for this type of object
    IMMUTABLE = False  # If set to true then the object is decoded straight away
    INJECT_CREATION_TRACKING = False
    # The latest migration, if there is one
    LATEST_MIGRATION = None  # type: migrations.ObjectMigration

    def __init__(self):
        assert self.TYPE is not None, "Must set the TYPE to a type of or a tuple of types"
        if self.INJECT_CREATION_TRACKING:
            inject_creation_tracking(self.TYPE)

    def new(self, encoded_saved_state):  # pylint: disable=unused-argument
        """Create a new blank object of this type"""
        cls = self.TYPE
        return cls.__new__(cls)

    @abstractmethod
    def yield_hashables(self, obj, hasher):
        """Produce a hash representing the value"""

    @abstractmethod
    def eq(self, one, other) -> bool:  # pylint: disable=invalid-name
        """Determine if two objects are equal"""

    @abstractmethod
    def save_instance_state(self, obj, saver):
        """Save the instance state of an object, should return a saved instance"""

    @abstractmethod
    def load_instance_state(self, obj, saved_state, loader: 'mincepy.Loader'):
        """Take the given blank object and load the instance state into it"""

    def get_version(self) -> Optional[int]:
        """Gets the version of the latest migration, returns None if there is not migration"""
        if self.LATEST_MIGRATION is None:
            return None

        version = self.LATEST_MIGRATION.VERSION
        if version is None:
            raise RuntimeError("Object '{}' has a migration ({}) which has no version "
                               "number".format(self.TYPE, self.LATEST_MIGRATION))

        return version

    def ensure_up_to_date(self, saved_state, version: Optional[int], loader: 'mincepy.Loader'):
        """Apply any migrations that are necessary to this saved state.  If no migrations are
        necessary then None is returned"""
        latest_version = None if self.LATEST_MIGRATION is None else self.LATEST_MIGRATION.VERSION
        if latest_version == version:
            return None

        if latest_version is None or (version is not None and latest_version < version):
            raise exceptions.VersionError(
                "This codebase's version of '{}' is older ({}) than the saved version ({}).  Check "
                "for updates.".format(pytray.pretty.type_string(self.TYPE), latest_version,
                                      version))

        to_apply = self._get_migrations(version)
        if not to_apply:
            return None

        total = len(to_apply)
        logger.info("Migrating saved state of '%s' from version %s to %i (%i migrations to apply)",
                    pytray.pretty.type_string(self.TYPE), version, self.get_version(), total)
        for i, migration in enumerate(to_apply):
            saved_state = migration.upgrade(saved_state, loader)
            logger.info("Migration '%s' applied (%i/%i)", pytray.pretty.type_string(migration),
                        i + 1, total)

        logger.info("Migration of '%s' completed successfully",
                    pytray.pretty.type_string(self.TYPE))

        return saved_state

    def _get_migrations(self, version: Optional[int]) -> Sequence[migrations.ObjectMigration]:
        """Get the sequence of migrations that needs to be applied to a given version"""
        if self.LATEST_MIGRATION is None:
            return []  # No migrations we can apply

        to_apply = []
        current = self.LATEST_MIGRATION
        while version is None or version < current.VERSION:
            to_apply.append(current)
            current = current.PREVIOUS
            if current is None:
                break

        to_apply.reverse()
        return to_apply


class BaseHelper(TypeHelper, metaclass=ABCMeta):
    """A base helper that defaults to yielding hashables directly on the object
    and testing for equality using == given two objects.  This behaviour is fairly
    standard and therefor more type helpers will want to subclass from this class."""

    # pylint: disable=abstract-method

    def yield_hashables(self, obj, hasher):
        yield from hasher.yield_hashables(obj)

    def eq(self, one, other) -> bool:
        return one == other


class WrapperHelper(TypeHelper):
    """Wraps up an object type to perform the necessary Historian actions"""

    # pylint: disable=invalid-name

    def __init__(self, obj_type: Type[types.SavableObject]):
        self.TYPE = obj_type
        self.TYPE_ID = obj_type.TYPE_ID
        self.LATEST_MIGRATION = obj_type.LATEST_MIGRATION
        super(WrapperHelper, self).__init__()

    def yield_hashables(self, obj, hasher):
        yield from self.TYPE.yield_hashables(obj, hasher)

    def eq(self, one, other) -> bool:
        return self.TYPE.__eq__(one, other)

    def save_instance_state(self, obj: types.Savable, saver):
        return self.TYPE.save_instance_state(obj, saver)

    def load_instance_state(self, obj, saved_state: types.Savable, loader):
        self.TYPE.load_instance_state(obj, saved_state, loader)



import os
from typing import Optional

import deprecation

from . import archive_factory
from . import historians
from . import plugins
from . import version

__all__ = ('connect', 'get_historian', 'set_historian', 'DEFAULT_ARCHIVE_URI', 'ENV_ARCHIVE_URI', \
           'archive_uri', 'load', 'save')

DEFAULT_ARCHIVE_URI = 'mongodb://localhost/mincepy'
ENV_ARCHIVE_URI = 'MINCEPY_ARCHIVE'
CURRENT_HISTORIAN = None


@deprecation.deprecated(deprecated_in="0.15.3",
                        removed_in="0.16.0",
                        current_version=version.__version__,
                        details="Use default_archive_uri() instead")
def archive_uri() -> Optional[str]:
    """Returns the default archive URI.  This is currently being taken from the environmental
    MINCEPY_ARCHIVE, however it may chance to include a config file in the future."""
    return os.environ.get(ENV_ARCHIVE_URI, DEFAULT_ARCHIVE_URI)


@deprecation.deprecated(deprecated_in="0.15.3",
                        removed_in="0.16.0",
                        current_version=version.__version__,
                        details="Use connect(set_global=False) instead")
def create_default_historian():
    """Create a default historian using the current `archive_uri()`"""
    uri = default_archive_uri()
    if uri:
        return archive_factory.create_historian(uri)

    return None


def connect(uri: str = '', use_globally=False) -> historians.Historian:
    """Connect to an archive and return a corresponding historian

    :param uri: the URI of the archive to connect to
    :param use_globally: if True sets the newly create historian as the current global historian
    """
    uri = uri or default_archive_uri()
    hist = archive_factory.create_historian(uri, apply_plugins=True)
    if use_globally:
        set_historian(hist, apply_plugins=False)
    return hist


def default_archive_uri() -> Optional[str]:
    """Returns the default archive URI.  This is currently being taken from the environmental
    MINCEPY_ARCHIVE, however it may chance to include a config file in the future."""
    return os.environ.get(ENV_ARCHIVE_URI, DEFAULT_ARCHIVE_URI)


# region Globals


def get_historian(create=True) -> Optional[historians.Historian]:
    """Get the currently set global historian.  If one doesn't exist and create is True then this
    call will attempt to create a new default historian using connect()"""
    global CURRENT_HISTORIAN  # pylint: disable=global-statement
    if CURRENT_HISTORIAN is None and create:
        # Try creating a new one
        connect(use_globally=True)

    return CURRENT_HISTORIAN


def set_historian(new_historian: Optional[historians.Historian], apply_plugins=True):
    """Set the current global historian.  Optionally load all plugins.
    To reset the historian pass None.
    """
    global CURRENT_HISTORIAN  # pylint: disable=global-statement
    if new_historian is not None and apply_plugins:
        new_historian.register_types(plugins.get_types())
    CURRENT_HISTORIAN = new_historian


def load(*obj_ids_or_refs):
    """Load one or more objects using the current global historian"""
    return get_historian().load(*obj_ids_or_refs)


def save(*objs):
    """Save one or more objects using the current global historian"""
    return get_historian().save(*objs)


# endregion



import bson

import mincepy

# pylint: disable=invalid-name

# Problem with subscripting archive, bug report here:
# https://github.com/PyCQA/pylint/issues/2822
_Archive = mincepy.Archive[bson.ObjectId]  # pylint: disable=unsubscriptable-object
SnapshotId = _Archive.SnapshotId



"""This module defines the data record and other objects and functions related to storing things
in an archive."""

import copy
from collections import namedtuple
import datetime
import typing
from typing import Optional, Iterable, Sequence, Union, Tuple, Any, Mapping
import uuid

import pytray.tree

from . import utils

__all__ = 'OBJ_ID', 'TYPE_ID', 'CREATION_TIME', 'VERSION', 'STATE', 'SNAPSHOT_TIME', \
          'SNAPSHOT_HASH', 'EXTRAS', 'ExtraKeys', 'DELETED', 'DataRecord', 'SnapshotRef', \
          'DataRecordBuilder', 'StateSchema', 'SnapshotId'

OBJ_ID = 'obj_id'
TYPE_ID = 'type_id'
CREATION_TIME = 'creation_time'
VERSION = 'version'
STATE = 'state'
STATE_TYPES = 'state_types'
SNAPSHOT_HASH = 'snapshot_hash'
SNAPSHOT_TIME = 'snapshot_time'
EXTRAS = 'extras'

SchemaEntry = namedtuple('SchemaEntry', 'type_id version')


class ExtraKeys:
    # pylint: disable=too-few-public-methods
    CREATED_BY = '_created_by'  # The ID of the process the data was created in
    COPIED_FROM = '_copied_from'  # The reference to the snapshot that this object was copied from
    USER = '_user'  # The user that saved this snapshot
    HOSTNAME = '_hostname'  # The hostname of the computer this snapshot was saved on


DELETED = '!!deleted'  # Special state to denote a deleted record

IdT = typing.TypeVar('IdT')  # The archive ID type

#: The type ID - this is typically a UUID but can be something else in different contexts
TypeId = Any  # pylint: disable=invalid-name
#: A path to a field in in the record.  This is used when traversing a series of containers that can
#: be either dictionaries or lists and are therefore indexed by strings or integers
EntryPath = Sequence[Union[str, int]]
#: Type that represents a path to an entry in the record state and the corresponding type id
EntryInfo = Tuple[EntryPath, TypeId]


class SnapshotId(typing.Generic[IdT]):
    """A snapshot id identifies a particular version of an object (and the corresponding record), it
    it therefore composed of the object id and the version number."""

    #: The type id for references
    TYPE_ID = uuid.UUID('633c7035-64fe-4d87-a91e-3b7abd8a6a28')

    def __init__(self, obj_id, version: int):
        """Create a snapshot id by passing an object id and version"""
        super().__init__()
        self._obj_id = obj_id
        self._version = version

    def __str__(self):
        return "{}#{}".format(self._obj_id, self._version)

    def __repr__(self):
        return "SnapshotId({}, {})".format(self.obj_id, self.version)

    def __hash__(self):
        return (self.obj_id, self.version).__hash__()

    def __eq__(self, other):
        if not isinstance(other, SnapshotId):
            return False

        return self.obj_id == other.obj_id and self.version == other.version

    @property
    def obj_id(self) -> IdT:
        return self._obj_id

    @property
    def version(self) -> int:
        return self._version

    def to_dict(self) -> dict:
        r"""Convenience function to get a dictionary representation.
        Can be passed to constructor as \*\*kwargs"""
        return {'obj_id': self.obj_id, 'version': self.version}


SnapshotRef = SnapshotId


class DataRecord(
        namedtuple(
            'DataRecord',
            (
                # Object properties
                OBJ_ID,  # The ID of the object (spanning all snapshots)
                TYPE_ID,  # The type ID of this object
                CREATION_TIME,  # The time this object was created
                # Snapshot properties
                VERSION,  # The ID of this particular snapshot of the object
                STATE,  # The saved state of the object
                STATE_TYPES,  # The historian types saved in the state
                SNAPSHOT_HASH,  # The hash of the state
                SNAPSHOT_TIME,  # The time this snapshot was created
                EXTRAS,  # Additional data stored with the snapshot
            ))):
    """An immutable record that describes a snapshot of an object"""

    @classmethod
    def defaults(cls) -> dict:
        """Returns a dictionary of default values, the caller owns the dict and is free to modify
         it"""
        return {
            CREATION_TIME: None,
            SNAPSHOT_TIME: None,
            EXTRAS: {},
        }

    @classmethod
    def new_builder(cls, **kwargs) -> 'DataRecordBuilder':
        """Get a builder for a new data record, the version will be set to 0"""
        values = cls.defaults()
        values.update({
            CREATION_TIME: utils.DefaultFromCall(datetime.datetime.now),
            VERSION: 0,
            SNAPSHOT_TIME: utils.DefaultFromCall(datetime.datetime.now),
        })
        values.update(kwargs)
        return DataRecordBuilder(cls, values)

    @property
    def created_by(self) -> IdT:
        """Convenience property to get the creator from the extras"""
        return self.get_extra(ExtraKeys.CREATED_BY)

    def is_deleted_record(self) -> bool:
        """Does this record represent the object having been deleted"""
        return self.state == DELETED

    @property
    def snapshot_id(self) -> SnapshotId:
        """The snapshot id for this record"""
        return SnapshotId(self.obj_id, self.version)

    def get_copied_from(self) -> Optional[SnapshotId]:
        """Get the reference of the data record this object was originally copied from"""
        obj_ref = self.get_extra(ExtraKeys.COPIED_FROM)
        if obj_ref is None:
            return None

        return SnapshotId(**obj_ref)

    def get_extra(self, name):
        """Convenience function to get an extra from the record, returns None if the extra doesn't
        exist"""
        return self.extras.get(name, None)

    def child_builder(self, **kwargs) -> 'DataRecordBuilder':
        """
        Get a child builder from this DataRecord instance.  The following attributes will be copied
        over:

        * obj_id
        * type_id
        * creation_time
        * created_by

        and version will be incremented by one.
        """
        defaults = self.defaults()
        defaults.update({
            OBJ_ID: self.obj_id,
            TYPE_ID: self.type_id,
            CREATION_TIME: self.creation_time,
            VERSION: self.version + 1,
            SNAPSHOT_TIME: utils.DefaultFromCall(datetime.datetime.now),
            EXTRAS: copy.deepcopy(self.extras),
        })
        defaults.update(kwargs)
        return DataRecordBuilder(type(self), defaults)

    def get_references(self) -> Iterable[Tuple[EntryPath, SnapshotId]]:
        """Get all the references to other objects contained in this record"""
        references = []
        if self.state_types is not None and self.state is not None:
            for entry_info in filter(lambda entry: entry[1] == SnapshotId.TYPE_ID,
                                     self.state_types):
                path = entry_info[0]
                sid_info = pytray.tree.get_by_path(self.state, path)
                if sid_info is not None:
                    sid = SnapshotId(**sid_info)
                    references.append((path, sid))
        return references

    def get_state_schema(self) -> 'StateSchema':
        """Get the schema for the state.  This contains the types and versions for each member of
        the state"""
        schema = {}
        for entry in self.state_types:
            path = tuple(entry[0])
            type_id = entry[1]
            version = None
            if len(entry) > 2:
                version = entry[2]
            schema[path] = SchemaEntry(type_id, version)

        return schema


# State schema is a mapping where they key is a tuple containing a path to the entry in the schema
# and the value is information about that entry, including its type
# The reason a tuple used (instead of, say, a string) is that this way we can preserve the python
# type of they key e.g. (str, int, int str) could be used to reference a dictionary, then index in
# list, then index in list and then a string in a dictionary.
StateSchema = Mapping[tuple, SchemaEntry]
DataRecordBuilder = utils.NamedTupleBuilder[DataRecord]  # pylint: disable=invalid-name


def make_deleted_builder(last_record: DataRecord) -> DataRecordBuilder:
    """Get a record that represents the deletion of this object"""
    return last_record.child_builder(state=DELETED, state_types=None, snapshot_hash=None)



import pytest

import mincepy
from mincepy import testing

# pylint: disable=invalid-name


def test_delete(historian: mincepy.Historian):
    """Test deleting and then attempting to load an object"""
    car = testing.Car('lada')
    car_id = historian.save(car)
    historian.delete(car)
    with pytest.raises(mincepy.NotFound):
        historian.load(car_id)

    records = historian.history(car_id, as_objects=False)
    assert len(records) == 2, "There should be two record, the initial and the delete"
    assert records[-1].is_deleted_record()


def test_delete_from_obj_id(historian: mincepy.Historian):
    """Test deleting an object using it's object id"""
    car = testing.Car('skoda')
    car_id = car.save()
    del car

    historian.delete(car_id)

    with pytest.raises(mincepy.NotFound):
        historian.load(car_id)


def test_delete_in_transaction(historian: mincepy.Historian):
    saved_outside = testing.Car('fiat')
    outside_id = saved_outside.save()

    with historian.transaction():
        saved_inside = testing.Car('bmw')
        inside_id = saved_inside.save()
        historian.delete(inside_id)
        historian.delete(outside_id)

        with pytest.raises(mincepy.ObjectDeleted):
            historian.get_obj(obj_id=inside_id)
        with pytest.raises(mincepy.ObjectDeleted):
            historian.get_obj(obj_id=outside_id)

        with pytest.raises(mincepy.ObjectDeleted):
            historian.load(inside_id)
        with pytest.raises(mincepy.ObjectDeleted):
            historian.load(outside_id)

    with pytest.raises(mincepy.NotFound):
        historian.get_obj(obj_id=inside_id)
    with pytest.raises(mincepy.NotFound):
        historian.get_obj(obj_id=outside_id)

    with pytest.raises(mincepy.NotFound):
        historian.load(inside_id)
    with pytest.raises(mincepy.NotFound):
        historian.load(outside_id)


def test_delete_find(historian: mincepy.Historian):
    car = testing.Car('trabant')
    car_id = car.save()

    historian.delete(car_id)
    assert len(tuple(historian.find(obj_id=car_id))) == 0

    # Now check the archive
    assert len(tuple(historian.archive.find(obj_id=car_id))) == 2
    assert len(tuple(historian.archive.find(obj_id=car_id, state=mincepy.DELETED))) == 1


def test_delete_multiple_versions(historian: mincepy.Historian):
    car = testing.Car('skoda', 'green')
    car.save()
    car.colour = 'red'
    car.save()

    with historian.transaction():
        historian.delete(car)


def test_delete_twice(historian: mincepy.Historian):
    car = testing.Car('trabant')
    car_id = car.save()
    historian.delete(car_id)

    with pytest.raises(mincepy.NotFound):
        historian.delete(car_id)

    with pytest.raises(mincepy.NotFound):
        historian.delete(car)


def test_delete_twice_in_transaction(historian: mincepy.Historian):
    car = testing.Car('trabant')
    car_id = car.save()

    with historian.transaction():
        historian.delete(car_id)

        with pytest.raises(mincepy.NotFound):
            historian.delete(car_id)

        with pytest.raises(mincepy.NotFound):
            historian.delete(car)


def test_delete_referenced_by(historian: mincepy.Historian):
    car = testing.Car()
    person = testing.Person('martin', 35, car)
    person.save()

    with pytest.raises(mincepy.ReferenceError):
        historian.delete(car)

    # Check you can delete it in a transaction
    with historian.transaction():
        historian.delete(car)
        historian.delete(person)



# pylint: disable=unused-import
import random

import pytest

import mincepy
from mincepy.testing import archive_uri, mongodb_archive, historian, Car, Person, Garage
from . import utils


@pytest.fixture
def large_dataset(historian: mincepy.Historian):
    with historian.transaction():
        # Put in some cars
        Car('ferrari', 'red').save()
        Car('honda', 'white').save()
        Car('fiat', 'green').save()

        # Put in some yellow cars
        for _ in range(100):
            Car(make='renault', colour='yellow').save()

        # Put in some purely random
        for _ in range(100):
            Car(make=utils.random_str(5), colour=utils.random_str(3)).save()

        # Put in some people
        Person('sonia', 30).save()
        Person('martin', 35).save()
        Person('gavin', 34).save()
        Person('upul', 35).save()

        # Put in some with the same age
        for _ in range(100):
            Person(name=utils.random_str(5), age=54).save()

        # Put in some with the same name
        for _ in range(100):
            Person(name='mike', age=random.randint(0, 100)).save()

        # Put in some purely random
        for _ in range(100):
            Person(name=utils.random_str(5), age=random.randint(0, 100)).save()



from typing import Iterator, Sequence, Iterable

import mincepy  # pylint: disable=unused-import
from . import depositors
from . import helpers
from . import qops
from . import records

__all__ = ('Migrations',)


def _state_types_migration_condition(helper: helpers.TypeHelper) -> dict:
    return qops.elem_match_(
        **{
            '1':
                helper.TYPE_ID,  # Type id has to match, and,
            **qops.or_(
                # version has to be less than the current, or,
                {'2': qops.lt_(helper.get_version())},
                # there is no version number
                {'2': None})
        })


class Migrations:
    """The historian migrations namespace"""

    def __init__(self, historian: 'mincepy.Historian'):
        self._historian = historian

    def find_migratable_records(self) -> Iterator[records.DataRecord]:
        """Find archive records that can be migrated"""
        type_registry = self._historian.type_registry
        # Find all the types in the registry that have migrations
        have_migrations = [
            helper for helper in type_registry.type_helpers.values()
            if helper.get_version() is not None
        ]

        if not have_migrations:
            return []

        # Now, let's look for those records that would need migrating
        archive = self._historian.archive
        query = qops.or_(*list(map(_state_types_migration_condition, have_migrations)))
        return archive.find(state_types=query)

    def migrate_all(self) -> Sequence[records.DataRecord]:
        """Migrate all records that can be updated"""
        to_migrate = self.find_migratable_records()
        return self.migrate_records(to_migrate)

    def migrate_records(self,
                        to_migrate: Iterable[records.DataRecord]) -> Sequence[records.DataRecord]:
        """Migrate the given records (if possible).  Returns all the records that were actually
        migrated."""
        migrator = depositors.Migrator(self._historian)
        return migrator.migrate_records(to_migrate)



"""This module contains the names of the keys in the various collections used by the mongo
archive and methods to convert mincepy types to mongo collection entries and back"""
import functools

from bidict import bidict
import bson

import mincepy.records

SETTINGS_COLLECTION = 'settings'
GLOBAL_SETTINGS = 'global'

# region Data collection
OBJ_ID = 'obj_id'
VERSION = 'ver'
TYPE_ID = mincepy.records.TYPE_ID
CREATION_TIME = 'ctime'
STATE = 'state'
STATE_TYPES = 'state_types'
SNAPSHOT_HASH = 'hash'
SNAPSHOT_TIME = 'stime'
EXTRAS = mincepy.records.EXTRAS
REFERENCES = 'refs'

# Here we map the data record property names onto ones in our entry format.
# If a record property doesn't appear here it means the name says the same
KEY_MAP = bidict({
    mincepy.records.OBJ_ID: OBJ_ID,
    mincepy.records.TYPE_ID: TYPE_ID,
    mincepy.records.CREATION_TIME: CREATION_TIME,
    mincepy.records.VERSION: VERSION,
    mincepy.records.STATE: STATE,
    mincepy.records.STATE_TYPES: STATE_TYPES,
    mincepy.records.SNAPSHOT_HASH: SNAPSHOT_HASH,
    mincepy.records.SNAPSHOT_TIME: SNAPSHOT_TIME,
    mincepy.records.EXTRAS: EXTRAS,
})

# endregion


def to_record(entry) -> mincepy.DataRecord:
    """Convert a MongoDB data collection entry to a DataRecord"""
    record_dict = mincepy.DataRecord.defaults()

    record_dict[mincepy.OBJ_ID] = entry[OBJ_ID]
    record_dict[mincepy.VERSION] = entry[VERSION]

    # Invert our mapping of keys back to the data record property names and update over any
    # defaults
    record_dict.update(
        {recordkey: entry[dbkey] for recordkey, dbkey in KEY_MAP.items() if dbkey in entry})

    return mincepy.DataRecord(**record_dict)


@functools.singledispatch
def to_document(record, exclude_defaults=False) -> dict:
    """Convert mincepy record information to a MongoDB document.  Optionally exclude entries that
    have the same value as the default"""
    raise TypeError(record.__class__)


@to_document.register(mincepy.DataRecord)
def _(record: mincepy.DataRecord, exclude_defaults=False) -> dict:
    """Convert a DataRecord to a MongoDB document with our keys"""
    defaults = mincepy.DataRecord.defaults()
    entry = {}
    for key, item in record._asdict().items():
        db_key = KEY_MAP[key]
        # Exclude entries that have the default value
        if not (exclude_defaults and key in defaults and defaults[key] == item):
            entry[db_key] = item

    return entry


@to_document.register(dict)
def _(record: dict, exclude_defaults=False) -> dict:
    """Convert a dictionary containing record keys to a MongoDB document with our keys"""
    defaults = mincepy.DataRecord.defaults()
    entry = {}
    for key, item in record.items():
        db_key = KEY_MAP[key]
        # Exclude entries that have the default value
        if not (exclude_defaults and key in defaults and defaults[key] == item):
            entry[db_key] = item

    return entry


def remap(record_dict: dict) -> dict:
    """Given a dictionary return a new dictionary with the key names that we use"""
    remapped = {}
    for key, value in record_dict.items():
        split_key = key.split('.')
        base = KEY_MAP[split_key[0]]
        split_key[0] = base
        remapped['.'.join(split_key)] = value
    return remapped


def to_id_dict(sid: mincepy.SnapshotId) -> dict:
    return {OBJ_ID: sid.obj_id, VERSION: sid.version}


def sid_from_dict(record: dict):
    return mincepy.SnapshotId(record[OBJ_ID], record[VERSION])


def sid_from_str(sid_str: str):
    parts = sid_str.split('#')
    return mincepy.SnapshotId(bson.ObjectId(parts[0]), int(parts[1]))



import logging

from . import historians
from . import plugins

__all__ = 'create_archive', 'create_historian'

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def create_archive(uri: str):
    """Create an archive type based on a uri string"""
    archive = None
    if uri.startswith('mongodb'):
        from . import mongo  # pylint: disable=import-outside-toplevel
        archive = mongo.connect(uri)

    if archive is None:
        raise ValueError("Unknown archive string: {}".format(uri))

    logger.info('Connected to archive with uri: %s', uri)
    return archive


def create_historian(archive_uri: str, apply_plugins=True) -> historians.Historian:
    """Convenience function to create a standard historian directly from an archive URI"""
    historian = historians.Historian(create_archive(archive_uri))
    if apply_plugins:
        historian.register_types(plugins.get_types())

    return historian



import gc
import uuid

import pytest

import mincepy
import mincepy.builtins
from mincepy.testing import Car, Garage, Cycle

# pylint: disable=invalid-name, too-few-public-methods


def test_basic_save_load(historian: mincepy.Historian):
    car = Car('nissan', 'white')

    car_id = historian.save(car)
    del car
    loaded_car = historian.load(car_id)

    assert loaded_car.make == 'nissan'
    assert loaded_car.colour == 'white'


def test_nested_references(historian: mincepy.Historian):
    car = Car()
    garage = Garage(car)

    car_id = historian.save(car)
    garage_id = historian.save(garage)

    # Now change the car
    car.make = 'fiat'
    car.colour = 'white'

    historian.save(car)
    # Try loading while the object is still alive
    loaded_garage = historian.load(garage_id)
    assert loaded_garage is garage

    # Now delete and load
    del garage
    del car
    loaded_garage2 = historian.load(garage_id)
    # Should be the last version fo the car
    assert loaded_garage2.car.make == 'fiat'

    assert len(historian.history(car_id)) == 2
    # The following may seem counter intuitive that we only have one history
    # entry for garage.  But above we only saved it once.  It's just that when
    # we load the garage again we get the 'latest' version it's contents i.e.
    # the newer version of the car
    assert len(historian.history(garage_id)) == 1


def test_create_delete_load(historian: mincepy.Historian):
    car = Car('honda', 'red')
    car_id = historian.save(car)
    del car

    loaded_car = historian.load(car_id)
    assert loaded_car.make == 'honda'
    assert loaded_car.colour == 'red'


# def test_save_as(historian: mincepy.Historian):
#     """Check the save_as functionality in historian"""
#     car = Car('ferrari')
#
#     car_id = historian.save(car)
#     assert historian.get_current_record(car) is not None
#
#     # Now create a new car and save that over the old one
#     new_car = Car('honda')
#
#     new_car_id = historian.save_as(new_car, car_id)
#     assert car_id == new_car_id
#
#     with pytest.raises(mincepy.NotFound):
#         # Now that 'old' car should not be known to the historian
#         historian.get_current_record(car)
#
#     assert historian.get_current_record(new_car) is not None


def test_type_helper(historian: mincepy.Historian):
    """Check that a type helper can be used to make a non-historian compatible type compatible"""

    class Bird:

        def __init__(self, specie='hoopoe'):
            self.specie = specie

    class BirdHelper(mincepy.TypeHelper):
        TYPE = Bird
        TYPE_ID = uuid.UUID('5cc59e03-ea5d-43ff-8814-3b6f2e22cd76')

        def yield_hashables(self, obj, hasher):
            yield from hasher.yield_hashables(obj.specie)

        def eq(self, one, other) -> bool:
            return one.specie == other.specie

        def save_instance_state(self, obj, _depositor):
            return obj.specie

        def load_instance_state(self, obj, saved_state, _depositor):
            obj.specie = saved_state

    bird = Bird()
    with pytest.raises(TypeError):
        historian.save(bird)

    # Now register the helper...
    historian.register_type(BirdHelper())
    # ...and we should be able to save
    assert historian.save(bird) is not None


def test_storing_internal_object(historian: mincepy.Historian):

    class Person(mincepy.SavableObject):
        TYPE_ID = uuid.UUID('f6f83595-6375-4bc4-89f2-d8f31a1286b0')

        def __init__(self, car):
            super(Person, self).__init__()
            self.car = car  # This person 'owns' the car

        def __eq__(self, other):
            return self.car == other.car

        def yield_hashables(self, hasher):
            yield from hasher.yield_hashables(self.car)

        def save_instance_state(self, _depositor):
            return {'car': self.car}

        def load_instance_state(self, saved_state, _depositor):
            self.car = saved_state['car']

    ferrari = Car('ferrari')
    mike = Person(ferrari)

    mike_id = historian.save(mike)
    del mike

    loaded_mike = historian.load(mike_id)
    assert loaded_mike.car.make == 'ferrari'

    # Default is to save by value so two cars should not be the same
    assert loaded_mike.car is not ferrari
    assert loaded_mike.car == ferrari  # But values should match


def test_load_unknown_object(mongodb_archive, historian: mincepy.Historian):
    """Make up an ID an try to load it"""
    obj_id = mongodb_archive.create_archive_id()
    with pytest.raises(mincepy.NotFound):
        historian.load(obj_id)


def test_cyclic_ref_simple(historian: mincepy.Historian):
    a = Cycle()
    a.ref = a  # Cycle complete
    a_id = historian.save(a)
    del a
    gc.collect()  # Force a garbage collection

    loaded_a = historian.load(a_id)
    assert loaded_a.ref is loaded_a


def test_cyclic_ref_complex(historian: mincepy.Historian):
    a = Cycle()
    b = Cycle(a)
    a.ref = b  # Cycle complete

    a_id = historian.save(a)
    del a

    loaded_a = historian.load(a_id)
    assert loaded_a.ref is b
    assert b.ref is loaded_a

    del loaded_a, b
    gc.collect()  # Force a garbage collection
    reloaded_a = historian.load(a_id)
    assert isinstance(reloaded_a.ref, Cycle)


def test_transaction_rollback(historian: mincepy.Historian):
    ferrari = Car('ferrari')
    with historian.transaction() as trans:
        # Within the transaction should be able to save and load
        ferrari_id = historian.save(ferrari)
        loaded = historian.load(ferrari_id)
        assert loaded is ferrari
        trans.rollback()

    # But now that I rolled back the object should not be available
    with pytest.raises(mincepy.NotFound):
        historian.load(ferrari_id)

    ferrari_id = historian.save(ferrari)
    assert historian.load(ferrari_id) is ferrari


def test_replace_simple(historian: mincepy.Historian):

    def paint_shop(car, colour):
        """An imaginary function that modifies an object but returns a copy rather than an in
        place modification"""
        return Car(car.make, colour)

    honda = Car('honda', 'yellow')
    honda_id = historian.save(honda)

    # Now paint the honda
    new_honda = paint_shop(honda, 'green')
    assert historian.get_obj_id(honda) == honda_id

    # Now we know that this is a 'continuation' of the history of the original honda, so replace
    historian.replace(honda, new_honda)
    assert historian.get_obj_id(honda) is None

    assert historian.get_obj_id(new_honda) == honda_id
    historian.save(new_honda)
    del honda, new_honda

    loaded = historian.load(honda_id)
    assert loaded.make == 'honda'
    assert loaded.colour == 'green'


def test_replace_invalid(historian: mincepy.Historian):
    honda = Car('honda', 'yellow')
    historian.save(honda)
    with pytest.raises(TypeError):
        historian.replace(honda, Garage())


def test_user_info(historian: mincepy.Historian):
    user_info = historian.get_user_info()
    assert user_info[mincepy.ExtraKeys.USER]
    assert user_info[mincepy.ExtraKeys.HOSTNAME]

    car = Car()
    historian.save(car)
    record = historian.get_current_record(car)

    assert record.extras[mincepy.ExtraKeys.USER] == user_info[mincepy.ExtraKeys.USER]
    assert record.extras[mincepy.ExtraKeys.HOSTNAME] == user_info[mincepy.ExtraKeys.HOSTNAME]


def test_save_as_ref(historian: mincepy.Historian):

    class Person(mincepy.SimpleSavable):
        TYPE_ID = uuid.UUID('692429b6-a08b-489a-aa09-6eb3174b6405')
        ATTRS = (mincepy.AsRef('car'),)  # Save the car by reference

        def __init__(self, car):
            super().__init__()
            self.car = car

    car = Car()
    # Both martin and sonia have the same car
    martin = Person(car)
    sonia = Person(car)
    martin_id, sonia_id = historian.save(martin, sonia)
    del martin, sonia, car

    # No reload and check they still have the same car
    martin, sonia = historian.load(martin_id, sonia_id)
    assert martin.car is sonia.car


def test_encode_nested(historian: mincepy.Historian):

    class CarDelegate(mincepy.SimpleSavable):
        TYPE_ID = uuid.UUID('c0148a43-c0c0-4d2b-9262-ed1c8c6ab2fc')
        ATTRS = ('car',)

        def __init__(self, car):
            super(CarDelegate, self).__init__()
            self.car = car

        def save_instance_state(self, saver):
            return {'car': self.car}

        def load_instance_state(self, saved_state, _loader):
            self.car = saved_state['car']

    historian.register_type(CarDelegate)

    car = Car()
    delegate = CarDelegate(car)
    garage = Garage(delegate)

    garage_id = historian.save(garage)
    del garage

    historian.load(garage_id)



import abc
from typing import List, Type

import pymongo.database

from . import settings

VERSION = 'version'
MIGRATIONS = 'migrations'


class MigrationError(RuntimeError):
    """An error that occurred during migration"""


class Migration(metaclass=abc.ABCMeta):
    """Migration class to be used when making changes to the database.  The implementer should
    provide a version number that higher than the previous as well as a reference to the previous
    migration.

    Migrations should avoid using constants from the codebase as these could change.  Ideally each
    migration should be self-contained and not refer to any other parts of the code.
    """

    # pylint: disable=invalid-name
    NAME = None
    VERSION = None
    PREVIOUS = None

    def __init__(self):
        if self.NAME is None:
            self.NAME = self.__class__.__name__

        if self.VERSION is None:
            raise RuntimeError("Migration version not set")

    @abc.abstractmethod
    def upgrade(self, database: pymongo.database.Database):
        """Apply the transformations to bring the database up to this version"""
        current = settings.get_settings(database) or {}
        # Update the version
        current[VERSION] = self.VERSION
        # and migrations
        current.setdefault(MIGRATIONS, []).append(self.NAME)
        settings.set_settings(database, current)


class MigrationManager:

    def __init__(self, latest: Type[Migration]):
        self.latest = latest

    def migration_required(self, database: pymongo.database.Database) -> bool:
        return len(self._get_required_migrations(database)) > 0

    def migrate(self, database: pymongo.database.Database) -> int:
        """Migrate the database and returns the number of migrations applied"""
        migrations = self._get_required_migrations(database)

        if migrations:
            for migration in reversed(migrations):
                migration().upgrade(database)

            # Recheck the settings to make sure the migration has been applied
            current = settings.get_settings(database)
            if current[MIGRATIONS][-1] != migrations[0].NAME:
                raise RuntimeError(
                    "After applying migrations, latest migration doesn't seem to be applied")

        return len(migrations)

    def get_migration_sequence(self) -> List[Type[Migration]]:
        """Get the sequence of migrations in REVERSE order i.e. the 0th entry is the latest"""
        # Establish the sequence of migrations
        migrations = [self.latest]
        while migrations[-1].PREVIOUS is not None:
            migrations.append(migrations[-1].PREVIOUS)

        return migrations

    def _get_required_migrations(self,
                                 database: pymongo.database.Database) -> List[Type[Migration]]:
        current = settings.get_settings(database) or {}
        migrations = self.get_migration_sequence()
        # Find out where we're at
        applied = current.get(MIGRATIONS, [])

        if applied:
            # Find the last one that was applied
            while migrations and applied[-1] != migrations.pop().NAME:
                pass

        # Anything left in the migrations list needs to be applied
        return migrations


def ensure_up_to_date(database: pymongo.database.Database, latest: Type[Migration]):
    """Apply any necessary migrations to the database"""
    current = settings.get_settings(database) or {}
    current_version = current.get(VERSION, None)
    if current_version and current_version > latest.VERSION:
        raise MigrationError(
            "The current database version ({}) is higher than the code version ({}) you may need "
            "to update your version of the code".format(current_version, latest.VERSION))

    migrator = MigrationManager(latest)
    return migrator.migrate(database)



from mincepy import type_registry
from test.common import A, B, C

# pylint: disable=invalid-name


def test_get_version_info():
    registry = type_registry.TypeRegistry()
    registry.register_type(A)
    registry.register_type(B)
    registry.register_type(C)

    c_info = registry.get_version_info(C)
    # Check versions
    assert c_info[A.TYPE_ID] == A.LATEST_MIGRATION.VERSION
    assert c_info[B.TYPE_ID] == B.LATEST_MIGRATION.VERSION
    assert c_info[C.TYPE_ID] == C.LATEST_MIGRATION.VERSION


def test_automatic_registration_of_parent():
    """Test that if a historian type with historian type ancestor(s) is registered then so are its
    ancestors"""
    registry = type_registry.TypeRegistry()
    registry.register_type(C)
    assert A in registry
    assert B in registry



import uuid

import mincepy


def test_transaction_snapshots(historian: mincepy.Historian):

    class ThirdPartyPerson:
        """A class from a third party library"""

        def __init__(self, name):
            self.name = name

    class PersonHelper(mincepy.TypeHelper):
        TYPE_ID = uuid.UUID('62d8c767-14bc-4437-a9a3-ca5d0ce65d9b')
        TYPE = ThirdPartyPerson
        INJECT_CREATION_TRACKING = True

        def yield_hashables(self, obj, hasher):
            yield from hasher.yield_hashables(obj.name)

        def eq(self, one, other) -> bool:
            return one.name == other.name

        def save_instance_state(self, obj, saver):
            return obj.name

        def load_instance_state(self, obj, saved_state, loader):
            obj.name = saved_state.name

    person_helper = PersonHelper()
    historian.register_type(person_helper)

    person_maker = mincepy.Process('person maker')

    with person_maker.running():
        martin = ThirdPartyPerson('Martin')

    historian.save(martin)
    assert historian.created_by(martin) == historian.get_obj_id(person_maker)
